
__author__    = "Radical.Utils Development Team (Andre Merzky)"
__copyright__ = "Copyright 2013, RADICAL@Rutgers"
__license__   = "MIT"


# ------------------------------------------------------------------------------
#
'''
Parse (absolute and relative) URLs.

See RFC 1808: "Relative Uniform Resource Locators", by R. Fielding,
UC Irvine, June 1995.
'''

__all__ = ["urlparse", "urlunparse", "urljoin", "urldefrag",
           "urlsplit", "urlunsplit"]

# ------------------------------------------------------------------------------
#
# A classification of schemes ('' means apply by default)
#
uses_relative    = ['ftp',      'http',      'gopher',   'nntp',    'imap',
                    'wais',     'file',      'https',    'shttp',   'mms',
                    'prospero', 'rtsp',      'rtspu',    'sftp',    '']

uses_netloc      = ['ftp',      'http',      'gopher',   'nntp',    'telnet',
                    'imap',     'wais',      'file',     'mms',     'https',
                    'shttp',    'snews',     'prospero', 'rtsp',    'rtspu',
                    'rsync',    '',          'svn',      'svn+ssh', 'sftp',
                    'pbs',      'pbs+ssh',   'sge',      'sge+ssh', 'fork',
                    'ssh',      'xt5torque', 'ec2',      'euca',    'redis'
                    'pbs+gsissh',  'sge+gsissh', 'xt5torque+gsissh',
                    'xt5torque+ssh']

non_hierarchical = ['gopher', 'hdl',      'mailto',   'news',
                    'telnet', 'wais',     'imap',     'snews',  'sip',   'sips']

uses_params      = ['ftp',    'hdl',      'prospero', 'http',   'imap',
                    'https',  'shttp',    'rtsp',     'rtspu',  'sip',   'sips',
                    'mms',    '',         'sftp']

uses_query       = ['http',   'wais',     'imap',     'https',  'shttp', 'mms',
                    'gopher', 'rtsp',     'rtspu',    'sip',    'sips',  'srm',
                    '']

uses_fragment    = ['ftp',    'hdl',      'http',     'gopher', 'news',
                    'nntp',   'wais',     'https',    'shttp',  'snews',
                    'file',   'prospero', '']

uses_hash_in_hostname     = ['go']
preserve_case_in_hostname = ['go']

# Characters valid in scheme names
scheme_chars              = ('abcdefghijklmnopqrstuvwxyz'
                             'ABCDEFGHIJKLMNOPQRSTUVWXYZ'
                             '0123456789'
                             '+-.')


# ------------------------------------------------------------------------------
#
MAX_CACHE_SIZE = 20
_parse_cache   = dict()


# ------------------------------------------------------------------------------
#
def clear_cache():
    '''
    Clear the parse cache
    '''

    global _parse_cache                                  # pylint: disable=W0603
    _parse_cache = dict()


# ------------------------------------------------------------------------------
#
class BaseResult(tuple):
    '''
    Base class for the parsed result objects.

    This provides the attributes shared by the two derived result
    objects as read-only properties.  The derived classes are
    responsible for checking the right number of arguments were
    supplied to the constructor.
    '''

    __slots__ = list()

    # Attributes that access the basic components of the URL:
    @property
    def scheme(self):    return self[0]  or ''
    @property
    def netloc(self):    return self[1]
    @property
    def path(self):      return self[2]  or ''
    @property
    def query(self):     return self[-2] or ''
    @property
    def fragment(self):  return self[-1] or ''

    # Additional attributes that provide access to parsed-out portions
    # of the netloc:
    @property
    def username(self):
        netloc = self.netloc
        if "@" in netloc:
            userinfo = netloc.split("@", 1)[0]
            if ":" in userinfo:
                userinfo = userinfo.split(":", 1)[0]
            return userinfo
        return ''


    @property
    def password(self):
        netloc = self.netloc
        if "@" in netloc:
            userinfo = netloc.split("@", 1)[0]
            if ":" in userinfo:
                return userinfo.split(":", 1)[1]
        return ''


    @property
    def hostname(self):
        netloc = self.netloc
        if "@" in netloc:
            netloc = netloc.split("@", 1)[1]
        if ":" in netloc:
            netloc = netloc.split(":", 1)[0]
        if self.scheme in preserve_case_in_hostname:
            return netloc or ''
        else:
            return netloc.lower() or ''


    @property
    def port(self):
        netloc = self.netloc
        if "@" in netloc:
            netloc = netloc.split("@", 1)[1]
        if ":" in netloc:
            port = netloc.split(":", 1)[1]
            return int(port, 10)
        return None


# ------------------------------------------------------------------------------
#
class SplitResult(BaseResult):

    __slots__ = list()

    # --------------------------------------------------------------------------
    #
    def __new__(cls, scheme, netloc, path, query, fragment):
        return BaseResult.__new__(
            cls, (scheme, netloc, path, query, fragment))

    # --------------------------------------------------------------------------
    #
    def geturl(self):
        return urlunsplit(self)


# ------------------------------------------------------------------------------
#
class ParseResult(BaseResult):

    __slots__ = list()

    # --------------------------------------------------------------------------
    #
    def __new__(cls, scheme, netloc, path, params, query, fragment):
        return BaseResult.__new__(
            cls, (scheme, netloc, path, params, query, fragment))

    # --------------------------------------------------------------------------
    #
    @property
    def params(self):
        return self[3]

    # --------------------------------------------------------------------------
    #
    def geturl(self):
        return urlunparse(self)


# ------------------------------------------------------------------------------
#
def urlparse(url, scheme='', allow_fragments=True):
    '''
    Parse a URL into 6 components:

        <scheme>://<netloc>/<path>;<params>?<query>#<fragment>

    Return a 6-tuple: (scheme, netloc, path, params, query, fragment).

    Note that we don't break the components up in smaller bits
    (e.g. netloc is a single string) and we don't expand % escapes.
    '''

    scheme, netloc, url, query, fragment = \
            urlsplit(url, scheme, allow_fragments)

    if scheme in uses_params and ';' in url:
        url, params = _splitparams(url)
    else:
        params = ''

    return ParseResult(scheme, netloc, url, params, query, fragment)


# ------------------------------------------------------------------------------
#
def _splitparams(url):

    if '/'  in url:
        i = url.find(';', url.rfind('/'))
        if i < 0:
            return url, ''
    else:
        i = url.find(';')

    return url[:i], url[i + 1:]


# ------------------------------------------------------------------------------
#
def _splitnetloc(url, start=0, allow_hash_in_hostname=False):

    delim = len(url)   # position of end of domain part of url, default is end

    if allow_hash_in_hostname:
        separators = '/?'
    else:
        separators = '/?#'

    for c in separators:    # look for delimiters; the order is NOT important
        wdelim = url.find(c, start)        # find first of this delim
        if wdelim >= 0:                    # if found
            delim = min(delim, wdelim)     # use earliest delim position

    return url[start:delim], url[delim:]   # return (domain, rest)


# ------------------------------------------------------------------------------
#
def urlsplit(url, scheme='', allow_fragments=True):
    '''
    Parse a URL into 5 components:

        <scheme>://<netloc>/<path>?<query>#<fragment>

    Return a 5-tuple: (scheme, netloc, path, query, fragment).

    Note that we don't break the components up in smaller bits
    (e.g. netloc is a single string) and we don't expand % escapes.
    '''

    allow_fragments = bool(allow_fragments)

    key    = url, scheme, allow_fragments, type(url), type(scheme)
    cached = _parse_cache.get(key, None)

    if cached:
        return cached

    if len(_parse_cache) >= MAX_CACHE_SIZE:  # avoid runaway growth
        clear_cache()

    netloc = query = fragment = ''
    i = url.find(':')

    if i > 0:

        if url[:i] == 'http':  # optimize the common case

            scheme = url[:i].lower()
            url    = url[i + 1:]

            if url[:2] == '//':
                netloc, url = _splitnetloc(url, 2)

            if allow_fragments and '#' in url:
                url, fragment = url.split('#', 1)

            if '?' in url:
                url, query = url.split('?', 1)

            v = SplitResult(scheme, netloc, url, query, fragment)
            _parse_cache[key] = v
            return v

        for c in url[:i]:
            if c not in scheme_chars:
                break
        else:
            scheme, url = url[:i].lower(), url[i + 1:]

    if url[:2] == '//':
        if scheme in uses_hash_in_hostname:
            netloc, url = _splitnetloc(url, 2, True)
        else:
            netloc, url = _splitnetloc(url, 2)

    if allow_fragments and scheme in uses_fragment and '#' in url:
        url, fragment = url.split('#', 1)

    if scheme in uses_query and '?' in url:
        url, query = url.split('?', 1)

    v = SplitResult(scheme, netloc, url, query, fragment)
    _parse_cache[key] = v
    return v


# ------------------------------------------------------------------------------
#
def urlunparse(data):
    '''
    Put a parsed URL back together again.  This may result in a
    slightly different, but equivalent URL, if the URL that was parsed
    originally had redundant delimiters, e.g. a ? with an empty query
    (the draft states that these are equivalent).
    '''

    (scheme, netloc, url, params, query, fragment) = data

    if params:
        url = "%s;%s" % (url, params)

    return urlunsplit((scheme, netloc, url, query, fragment))


# ------------------------------------------------------------------------------
#
def urlunsplit(data):

    (scheme, netloc, url, query, fragment) = data

    if netloc or (scheme and url[:2] != '//'):
        if url and url[:1] != '/':
            url = '/' + url
        url = '//' + (netloc or '') + url

    if scheme:   url = scheme + ':' + url
    if query:    url = url    + '?' + query
    if fragment: url = url    + '#' + fragment

    return url


# ------------------------------------------------------------------------------
#
def urljoin(base, url, allow_fragments=True):
    '''
    Join a base URL and a possibly relative URL to form an absolute
    interpretation of the latter.
    '''

    if not base:
        return url

    if not url:
        return base

    af = allow_fragments

    bscheme, bnetloc, bpath, bparams, bquery, _   = urlparse(base, '',      af)
    scheme, netloc, path, params, query, fragment = urlparse(url,  bscheme, af)

    if scheme != bscheme or scheme not in uses_relative:
        return url

    # if scheme in uses_netloc:
    if netloc:
        return urlunparse((scheme, netloc, path, params, query, fragment))

    netloc = bnetloc
    if path[:1] == '/':
        return urlunparse((scheme, netloc, path, params, query, fragment))

    if not (path or params or query):
        return urlunparse((scheme, netloc, bpath, bparams, bquery, fragment))

    segments = bpath.split('/')[:-1] + path.split('/')
    # FIXME: The stuff below is bogus in various ways...
    if segments[-1] == '.':
        segments[-1] = ''

    while '.' in segments:
        segments.remove('.')

    while 1:
        i = 1
        n = len(segments) - 1

        while i < n:
            if (segments[i] == '..' and segments[i - 1] not in ('', '..')):
                del segments[i - 1 : i + 1]
                break
            i = i + 1

        else:
            break

    if segments == ['', '..']:
        segments[-1] = ''

    elif len(segments) >= 2 and segments[-1] == '..':
        segments[-2:] = ['']

    return urlunparse((scheme, netloc, '/'.join(segments),
                       params, query, fragment))


# ------------------------------------------------------------------------------
#
def urldefrag(url):
    '''
    Removes any existing fragment from URL.

    Returns a tuple of the defragmented URL and the fragment.  If
    the URL contained no fragments, the second element is the
    empty string.
    '''

    if '#' in url:
        s, n, p, a, q, frag = urlparse(url)
        defrag = urlunparse((s, n, p, a, q, ''))
        return defrag, frag

    return url, ''


# ------------------------------------------------------------------------------





from .flux_service   import FluxService
from .flux_helper_v0 import FluxHelperV0 as _FluxHelperV0
from .flux_helper_v1 import FluxHelperV1 as _FluxHelperV1
from .flux_module    import FluxModule, spec_from_command, spec_from_dict

_fm = FluxModule()
if _fm.mode == 1: FluxHelper = _FluxHelperV1
else            : FluxHelper = _FluxHelperV0





from typing import Dict, Any

from ..typeddict import TypedDict
from ..serialize import to_msgpack, from_msgpack


# ------------------------------------------------------------------------------
#
class Message(TypedDict):

    # FIXME: register serialization methods for all message types

    _schema = {
        '_msg_type': str,
    }

    _defaults = {
        '_msg_type': None,
    }

    _msg_types = dict()


    # --------------------------------------------------------------------------
    #
    def _verify(self):
        assert self._msg_type


    @staticmethod
    def register_msg_type(msg_type, msg_class):
        Message._msg_types[msg_type] = msg_class


    @staticmethod
    def deserialize(data: Dict[str, Any]):

        msg_type = data.get('_msg_type')

        if msg_type is None:
            raise ValueError('no message type defined')

        if msg_type not in Message._msg_types:
            known = list(Message._msg_types.keys())
            raise ValueError('unknown message type [%s]: %s' % (msg_type, known))

        return Message._msg_types[msg_type](from_dict=data)


    def packb(self):
        return to_msgpack(self)

    @staticmethod
    def unpackb(bdata):
        return Message.deserialize(from_msgpack(bdata))


# ------------------------------------------------------------------------------





import re
import os

from .misc import ru_open

_pat = r'^\s*(?P<detail>(?P<short>[^-]*)(?P<base>-[^-@]+?)?(-[^@]+?)?(?P<branch>@.+?)?)\s*$'


# ------------------------------------------------------------------------------
#
# versioning mechanism:
#
#   - version_short :  1.2.3                       - used for installation
#   - version_detail:  1.2.3-v1.1-9-g0684b06@devel - used at runtime (ru.version)
#   - both are derived from the last git tag and branch information
#   - VERSION files are created on install, with the version_detail
#
def get_version(paths=None):
    """
    paths:
        a VERSION file containing the detailed version is checked for in every
        directory listed in paths.
    """

    if not paths:
        # by default, get version for myself
        pwd   = os.path.dirname(__file__)
        root  = "%s/.." % pwd
        paths = [root, pwd]

    if not isinstance(paths, list):
        paths = [paths]

    version_short  = None
    version_base   = None
    version_branch = None
    version_tag    = None
    version_detail = None
    err            = ''

    # if in any of the paths a VERSION file exists, we use the detailed version
    # in there.
    for path in paths:

        try:
            version_path = "%s/VERSION" % path

            with ru_open(version_path) as f:
                data    = f.read()
                lines   = data.split('\n')
                lines   = [line.strip() for line in lines]


                # make sure we have a valid version file
                assert len(lines) > 1

                version_short = lines[0]

                if len(lines) > 2:
                    version_base   = lines[1]
                    version_branch = lines[2]
                    version_tag    = lines[3]
                    version_detail = lines[4]

        except Exception as e:
            # ignore missing VERSION file, but keep error message
            err += '%s\n' % repr(e)

    # check if any one worked ok
    if not version_short:
        raise RuntimeError("Cannot determine version from %s (%s)"
                          % (paths, err.strip()))

    if version_detail is None:
        version_detail = version_short

    return (version_short, version_base, version_branch,
            version_tag, version_detail)


# ------------------------------------------------------------------------------





__author__    = "Radical.Utils Development Team (Andre Merzky)"
__copyright__ = "Copyright 2013, RADICAL@Rutgers"
__license__   = "MIT"


import threading

# pylint: disable=protected-access


# ------------------------------------------------------------------------------
#
def Lockable(cls):
    '''
    This class decorator will add lock/unlock methods to the thusly decorated
    classes, which will be enacted via an also added `threading.RLock` member
    (`self._rlock`)::

        @Lockable
        class A(object):

            def call(self):
                print 'locked: %s' % self._locked

    The class instance can then be used like this::

        a = A()
        a.call()
        a.lock()
        a.call()
        a.lock()
        a.call()
        a.unlock()
        a.call()
        with a:
          a.call()
        a.call()
        a.unlock()
        a.call()

    which will result in::

        locked: 0
        locked: 1
        locked: 2
        locked: 1
        locked: 2
        locked: 1
        locked: 0

    The class A can also internally use the lock, and can, for example, use::

        @Lockable
        class A(object):
            ...
            def work(self):
                with self:
                    # locked code section
                    ...
    '''

    if  hasattr(cls, '__enter__'):
        raise RuntimeError("Cannot make '%s' lockable -- has __enter__" % cls)

    if  hasattr(cls, '__exit__'):
        raise RuntimeError("Cannot make '%s' lockable -- has __exit__" % cls)

    if  hasattr(cls, '_rlock'):
        raise RuntimeError("Cannot make '%s' lockable -- has _rlock" % cls)

    if  hasattr(cls, '_locked'):
        raise RuntimeError("Cannot make '%s' lockable -- has _locked" % cls)

    if  hasattr(cls, 'locked'):
        raise RuntimeError("Cannot make '%s' lockable -- has locked" % cls)

    if  hasattr(cls, 'lock'):
        raise RuntimeError("Cannot make '%s' lockable -- has lock()" % cls)

    if  hasattr(cls, 'unlock'):
        raise RuntimeError("Cannot make '%s' lockable -- has unlock()" % cls)


    def locked(self):
        return self._locked

    def locker(self):
        self._rlock.acquire()
        self._locked += 1

    def unlocker(self, *args):                 # pylint: disable=unused-argument
        self._rlock.release()
        self._locked -= 1

    cls._rlock    = threading.RLock()
    cls._locked   = 0
    cls.locked    = locked
    cls.is_locked = locked
    cls.lock      = locker
    cls.unlock    = unlocker
    cls.__enter__ = locker
    cls.__exit__  = unlocker

    return cls


# ------------------------------------------------------------------------------
#
# @Lockable
# class A(object):
#
#     def call(self):
#         print 'locked 1: %s' % self.locked()
#
#         with self:
#             print 'locked 2: %s' % self.locked()
#
#         print 'locked 3: %s\n' % self.locked()
#
# print
# a = A()
# a.call()
# a.lock()
# a.call()
# a.lock()
# a.call()
# a.unlock()
# a.call()
# with a:
#   a.call()
# a.call()
# a.unlock()
# a.call()
# try:
#     a.unlock()
#     print 'oops\n'
# except:
#     print 'ok\n'
# a.lock()
# a.call()
# with a:
#   a.call()
# a.call()
# a.unlock()
# a.call()
# try:
#     a.unlock()
#     print 'oops\n'
# except:
#     print 'ok\n'

# ------------------------------------------------------------------------------






__author__    = "Radical.Utils Development Team (Andre Merzky)"
__copyright__ = "Copyright 2013, RADICAL@Rutgers"
__license__   = "MIT"


import os
import time
import uuid
import fcntl
import socket
import datetime
import threading

from .singleton import Singleton
from .misc      import dockerized, get_radical_base

TEMPLATE_SIMPLE  = "%(prefix)s.%(counter)04d"
TEMPLATE_UNIQUE  = "%(prefix)s.%(date)s.%(time)s.%(pid)06d.%(counter)04d"
TEMPLATE_PRIVATE = "%(prefix)s.%(host)s.%(user)s.%(days)06d.%(day_counter)04d"
TEMPLATE_UUID    = "%(prefix)s.%(uuid)s"


_cache = {'dir'       : list(),
          'user'      : None,
          'pid'       : os.getpid(),
          'dockerized': dockerized(),
          'rank'      : None}

if _cache['rank'] is None: _cache['rank'] = os.environ.get('PMIX_RANK')
if _cache['rank'] is None: _cache['rank'] = os.environ.get('PMI_RANK')
if _cache['rank'] is None: _cache['rank'] = os.environ.get('OMPI_COMM_WORLD_RANK')

if _cache['rank'] is None: _cache['rank'] = 0
else                     : _cache['rank'] = int(_cache['rank'])


# ------------------------------------------------------------------------------
#
class _IDRegistry(object, metaclass=Singleton):
    """
    This helper class (which is not exposed to any user of radical.utils)
    generates a sequence of continuous numbers for each known ID prefix.  It is
    a singleton, and thread safe (assuming that the Singleton metaclass supports
    thread safe construction).
    """

    # --------------------------------------------------------------------------
    def __init__(self):
        """
        Initialized the registry dict and the threading lock
        """

        self._rlock    = threading.RLock()
        self._registry = dict()


    # --------------------------------------------------------------------------
    def get_counter(self, prefix):
        """
        Obtain the next number in the sequence for the given prefix.
        If the prefix is not known, a new registry counter is created.
        """

        with self._rlock:

            if prefix not in self._registry:
                self._registry[prefix] = 0

            ret = self._registry[prefix]

            self._registry[prefix] += 1

        return ret


    # --------------------------------------------------------------------------
    def reset_counter(self, prefix, reset_all_others=False):
        """
        Reset the given counter to zero.
        """

        with self._rlock:

            if reset_all_others:
                # reset all counters *but* the one given
                for p in self._registry:
                    if p != prefix:
                        self._registry[p] = 0
            else:
                self._registry[prefix] = 0


# ------------------------------------------------------------------------------
#
# we create on private singleton instance for the ID registry.
#
_id_registry = _IDRegistry()
_BASE        = get_radical_base('utils')


# ------------------------------------------------------------------------------
#
ID_SIMPLE  = 'simple'
ID_UNIQUE  = 'unique'
ID_PRIVATE = 'private'
ID_CUSTOM  = 'custom'
ID_UUID    = 'uuid'


# ------------------------------------------------------------------------------
#
def generate_id(prefix: str, mode=ID_SIMPLE, ns=None):
    """
    Generate a human readable, sequential ID for the given prefix.

    The ID is by default very simple and thus very readable, but cannot be
    assumed to be globally unique -- simple ID uniqueness is only guaranteed
    within the scope of one python instance.

    If `mode` is set to the non-default type `ID_UNIQUE`, an attempt is made to
    generate readable but globally unique IDs -- although the level of
    confidence for uniqueness is significantly smaller than for, say UUIDs.

    The ID format per mode is:
    ID_SIMPLE  = "%(prefix)s.%(counter)04d"
    ID_UNIQUE  = "%(prefix)s.%(date)s.%(time)s.%(pid)06d.%(counter)04d"
    ID_PRIVATE = "%(prefix)s.%(host)s.%(user)s.%(days)06d.%(day_counter)04d"
    ID_UUID    = "%(prefix)s.%(uuid)s"

    Examples::

        print(radical.utils.generate_id('item'))
        print(radical.utils.generate_id('item'))
        print(radical.utils.generate_id('item', mode=radical.utils.ID_SIMPLE))
        print(radical.utils.generate_id('item', mode=radical.utils.ID_SIMPLE))
        print(radical.utils.generate_id('item', mode=radical.utils.ID_UNIQUE))
        print(radical.utils.generate_id('item', mode=radical.utils.ID_UNIQUE))
        print(radical.utils.generate_id('item', mode=radical.utils.ID_PRIVATE))
        print(radical.utils.generate_id('item', mode=radical.utils.ID_PRIVATE))
        print(radical.utils.generate_id('item', mode=radical.utils.ID_UUID))

    The above will generate the IDs:

        item.0000
        item.0001
        item.0002
        item.0003
        item.2014.07.30.13.13.44.0000
        item.2014.07.30.13.13.44.0001
        item.cameo.merzky.018375.0000
        item.cameo.merzky.018375.0001
        item.23cacb7e-0b08-11e5-9f0f-08002716eaa9

    where 'cameo' is the (short) hostname, 'merzky' is the username, and '18375'
    is 'days since epoch'.  The last element, the counter is unique for each id
    type and item type, and restarts for each session (application process).  In
    the last case though (`ID_PRIVATE`), the counter is reset for every new day,
    and can thus span multiple applications.

    'ns' argument can be specified to a value such that unique IDs are created
    local to that namespace. For example, you can create a session and use the
    session ID as a namespace for all the IDs of the objects of that execution.

    Example::

        sid  = generate_id('re.session', ID_PRIVATE)
        uid1 = generate_id('task.%(item_counter)04d', ID_CUSTOM, ns=sid)
        uid2 = generate_id('task.%(item_counter)04d', ID_CUSTOM, ns=sid)
        ...


    This will generate the following ids::

        re.session.rivendell.vivek.017548.0001
        task.0000
        task.0001

    where the `task.*` IDs are unique for the used sid namespace.

    The namespaces are stored under ```$RADICAL_BASE/.radical/utils/```.
    If `RADICAL_BASE` is not set, then `$HOME` is used.

    Note that for docker containers, we try to avoid hostname / username clashes
    and will, for `ID_PRIVATE`, revert to `ID_UUID`.
    """

    if not isinstance(prefix, str):
        raise TypeError('"prefix" must be a string, not %s' % type(prefix))

    if _cache['dockerized'] and mode == ID_PRIVATE:
        mode = ID_UUID

    if   mode == ID_CUSTOM : template = prefix
    elif mode == ID_UUID   : template = TEMPLATE_UUID
    elif mode == ID_SIMPLE : template = TEMPLATE_SIMPLE
    elif mode == ID_UNIQUE : template = TEMPLATE_UNIQUE
    elif mode == ID_PRIVATE: template = TEMPLATE_PRIVATE
    else: raise ValueError("unsupported mode '%s'", mode)

    return _generate_id(template, prefix, ns)


# ------------------------------------------------------------------------------
#
def _generate_id(template, prefix, ns=None):

    # FIXME: several of the vars below are constants, and many of them are
    #  rarely used in IDs. They should be created only once per module instance,
    #  and/or only if needed.

    state_dir = _BASE
    if ns:
        state_dir = os.path.join(_BASE, ns)

    if state_dir not in _cache['dir']:
        try   : os.makedirs(state_dir)
        except: pass
        _cache['dir'].append(state_dir)

    # seconds since epoch(float), and timestamp
    seconds = time.time()
    now     = datetime.datetime.fromtimestamp(seconds)
    days    = int(seconds / (60 * 60 * 24))

    if not _cache['user']:
        try:
            import getpass
            _cache['user'] = getpass.getuser()
        except:
            _cache['user'] = 'nobody'

    user = _cache['user']

    info = dict()

    info['day_counter' ] = 0
    info['item_counter'] = 0
    info['counter'     ] = 0
    info['prefix'      ] = prefix
    info['seconds'     ] = int(seconds)     # full seconds since epoch
    info['days'        ] = days             # full days since epoch
    info['user'        ] = user             # local username
    info['now'         ] = now
    info['date'        ] = "%04d.%02d.%02d" % (now.year, now.month,  now.day)
    info['time'        ] = "%02d.%02d.%02d" % (now.hour, now.minute, now.second)
    info['pid'         ] = _cache['pid']
    info['rank'        ] = _cache['rank']

    # the following ones are time consuming, and only done when needed
    if '%(host)' in template: info['host'] = socket.gethostname()  # localhost
    if '%(uuid)' in template: info['uuid'] = uuid.uuid1()          # plain uuid

    def _read_file_counter(name):
        output = 0
        try:
            fd = os.open(name, os.O_RDWR | os.O_CREAT)
            try:
                fcntl.flock(fd, fcntl.LOCK_EX)
            except OSError:
                # fcntl.flock might cause OSError: [Errno 524] Unknown error 524
                # (the case for Theta@ALCF)
                fcntl.lockf(fd, fcntl.LOCK_EX)
            os.lseek(fd, 0, os.SEEK_SET)
            data = os.read(fd, 256)
            if data: output = int(data)
            os.lseek(fd, 0, os.SEEK_SET)
            os.write(fd, str.encode("%d\n" % (output + 1)))
            os.close(fd)
        finally:
            try:
                os.close(fd)
            except:
                pass
        return output

    if '%(day_counter)' in template:
        fname = os.path.join(state_dir, 'ru_%s_%s.cnt' % (user, days))
        info['day_counter'] = _read_file_counter(fname)

    if '%(item_counter)' in template:

        # clean up "prefix" to use in file name
        #  FIXME: extend same procedure for other cases (with regex?)
        if '%(item_counter)' in prefix:
            prefix_parts = prefix.split('.')
            for _idx in range(len(prefix_parts)):
                if '%(item_counter)' in prefix_parts[_idx]:
                    prefix_parts[_idx] = 'item_counter'
                    break
            prefix = '.'.join(prefix_parts)

        fname = os.path.join(state_dir, 'ru_%s_%s.cnt' % (user, prefix))
        info['item_counter'] = _read_file_counter(fname)

    if '%(counter)' in template:
        info['counter'] = _id_registry.get_counter(prefix.replace('%', ''))

    try:
        ret = template % info

    except KeyError as e:
        raise ValueError('unknown pattern in template (%s)' % template) from e

    return ret


# ------------------------------------------------------------------------------
#
def reset_id_counters(prefix=None, reset_all_others=False):

    if not isinstance(prefix, list):
        prefix = [prefix]

    for p in prefix:
        if isinstance(p, str):
            p = p.replace('%', '')
        _id_registry.reset_counter(p, reset_all_others)

# ------------------------------------------------------------------------------



#!/usr/bin/env python 

import zmq

context = zmq.Context()

# Socket facing publishers
socket_in = context.socket(zmq.XSUB)
socket_in.bind("tcp://*:*")

# Socket facing Subscribers
socket_out = context.socket(zmq.XPUB)
socket_out.bind("tcp://*:*")

addr_in  = socket_in .getsockopt(zmq.LAST_ENDPOINT)
addr_out = socket_out.getsockopt(zmq.LAST_ENDPOINT)

print('PUB: %s' % addr_in)
print('SUB: %s' % addr_out)

with open('test.bridge.url', 'w') as fout:
    fout.write('PUB %s\n' % addr_in)
    fout.write('SUB %s\n' % addr_out)

zmq.proxy(socket_in, socket_out)

# We never get here...
socket_out.close()
socket_in.close()
context.term()





import time
import queue

import threading as mt

from functools   import partial
from collections import defaultdict
from typing      import List, Union

from ..misc       import as_list
from ..ids        import generate_id
from ..logger     import Logger

from .flux_module import FluxModule


# ------------------------------------------------------------------------------
#
class FluxHelperV0(object):

    # --------------------------------------------------------------------------
    #
    def __init__(self, uri : str,
                       log : Logger = None) -> None:

      # print('=== v0 flux helper ===')

        self._uri      = uri
        self._log      = log or Logger('radical.utils.flux')
        self._uid      = generate_id('ru.flux')

        self._fm       = FluxModule()
        self._handle   = self._fm.core.Flux(self._uri)
        self._api_lock = mt.Lock()
        self._exe      = self._fm.job.executor.FluxExecutor(
                                               handle_kwargs={'url': self._uri})

        self._idlock   = mt.Lock()          # lock ID dicts
        self._elock    = mt.Lock()          # lock event dict
        self._task_ids = dict()             # flux ID -> task ID
        self._flux_ids = dict()             # task ID -> flux ID
        self._events   = defaultdict(list)  # flux ID -> event list
        self._cbacks   = list()             # list of callbacks

        self._fm.verify()
        for line in (self._fm.version or '').splitlines():
            self._log.info('flux version: %s', line)


    # --------------------------------------------------------------------------
    #
    def start(self, launcher: str   = None) -> None:

        pass


    # --------------------------------------------------------------------------
    #
    def stop(self):

        with self._api_lock:

            # FIXME: shutdown flux instance
            self._uri          = None
            self._handle       = None


    # --------------------------------------------------------------------------
    #
    @property
    def uid(self) -> str:
        return self._uid

    @property
    def uri(self) -> str:
        return self._uri


    # --------------------------------------------------------------------------
    #
    def register_cb(self, cb: callable) -> None:

        with self._api_lock, self._elock:
            self._log.debug('register cb %s', cb)
            self._cbacks.append(cb)


    # --------------------------------------------------------------------------
    #
    def unregister_cb(self, cb: callable) -> None:

        with self._api_lock, self._elock:
            self._cbacks.remove(cb)


    # --------------------------------------------------------------------------
    #
    def _handle_events(self, fh   : 'flux.Flux',
                             fid  : 'flux.job.JobID',
                             event: 'flux.job.journal.JournalEvent' = None
                      ) -> None:

        self._log.debug('event for %s: %s', fid, event)
      # print('event %s: %s' % (fid, event))

        # if triggered by submit, check if we have anything to do
        if not event:
            if fid not in self._events:
              # print('no event')
                return

        # check if we can handle the event - otherwise store it
        if not self._cbacks:
          # print('no cbacks')
            self._events[fid].append(event)
            return

        # task is known, flush stored events
        for ev in self._events[fid]:
          # print('flush stored events')
            for cb in self._cbacks:
                try   : cb(fid, ev)
                except: self._log.exception('cb failed')
            self._events[fid] = []

        # process the current event
        if event:
          # print('process current event')
            for cb in self._cbacks:
                try   : cb(fid, event)
                except: self._log.exception('cb failed')


    # --------------------------------------------------------------------------
    #
    def submit(self, specs: List['flux.job.JobspecV1']) -> List[str]:

        with self._api_lock:

            if not self._handle:
                raise RuntimeError('flux instance not started')

            self._log.debug('== submit %d specs', len(specs))

            events = ['submit', 'depend', 'alloc', 'start', # 'cleanup',
                      'finish', 'release', 'free', 'clean', 'priority', 'exception']

            def event_cb(fid, fut, event):
                self._handle_events(self._handle, fid, event)

            futures = list()
            def id_cb(fut):
                flux_id = fut.jobid()
                idx     = fut.ru_idx
                for ev in events:
                    tmp_cb = partial(event_cb, flux_id)
                    fut.add_event_callback(ev, tmp_cb)
                futures.append([flux_id, idx, fut])
                self._log.debug('got flux id: %s: %s', idx, flux_id)

            for idx, spec in enumerate(specs):
                fut      = self._exe.submit(spec, waitable=True)
                fut.ru_idx = idx
                self._log.debug('%s: submitted  : %s', self._uid, idx)
                fut.add_jobid_callback(id_cb)

            # wait until we saw all jobid callbacks (assume 10 tasks/sec)
            timeout = len(specs)
            timeout = max(100, timeout)
            start   = time.time()
            self._log.debug('%s: wait %.2fsec for %d flux IDs',
                            self._uid, timeout, len(specs))
            while len(futures) < len(specs):
                time.sleep(0.1)
                self._log.debug('%s: wait %s / %s', self._uid,
                                                     len(futures), len(specs))
                if time.time() - start > timeout:
                    raise RuntimeError('%s: timeout on submission' % self._uid)
            self._log.info('got %d flux IDs', len(futures))

            # get flux_ids sorted by submission order (idx)
            flux_ids = [fut[0] for fut in sorted(futures, key=lambda x: x[1])]

            self._log.debug('%s: submitted: %s', self._uid, flux_ids)
            return flux_ids


    # --------------------------------------------------------------------------
    #
    def cancel(self, fids: [Union[str, List[str]]]) -> None:

        with self._api_lock:

            if not self._handle:
                raise RuntimeError('flux instance not started')

            with self._idlock:
                for fid in as_list(fids):
                    self._fm.job.cancel_async(self._handle, fid, reason='user cancel')


    # --------------------------------------------------------------------------
    #
    def wait(self, fids: [Union[str, List[str]]]) -> None:

        with self._api_lock:

            if not self._handle:
                raise RuntimeError('flux instance not started')

            for fid in fids:
                self._log.debug('wait  for %s', fid)
                self._fm.job.wait(self._handle, fid)


# ------------------------------------------------------------------------------





__author__    = "Radical.Utils Development Team"
__copyright__ = "Copyright 2013, RADICAL@Rutgers"
__license__   = "MIT"


import traceback

import threading as mt


_out_lock = mt.RLock()


# ------------------------------------------------------------------------------
#
# our futures have state, the states are defined here
#
# NOTE: these strings are carefully chosen to match the state specifiers of 
#       `radical.saga.Task` and `radical.saga.Job` instances.
#
NEW      = 'New'
RUNNING  = 'Running'
DONE     = 'Done'
FAILED   = 'Failed'
CANCELED = 'Canceled'

INITIAL  = [NEW]
FINAL    = [DONE, FAILED, CANCELED]


# ------------------------------------------------------------------------------
#
class Future(mt.Thread):
    """
    This `Future` class is a thin wrapper around Python's native `mt.Thread`
    class.  It is expected to wrap a callable, and to watch its execution.
    """

    # FIXME: we may want to use a thread pool to limit the number of threads


    # --------------------------------------------------------------------------
    #
    def __init__(self, call, *args, **kwargs):
        '''
        Construct the Future.  The first argument is expected to be a Python
        `callable` which is executed in its own thread.  All other arguments are
        passed blindly to that callable when `self.start()` is called.
        '''

        if not callable(call):
            raise ValueError("Thread requires a callable to function, not %s"
                            % (str(call)))

        mt.Thread.__init__(self)

        # NOTE: we use daemon threads to avoid termination issues
        self.daemon = True

        self._call      = call
        self._args      = args
        self._kwargs    = kwargs
        self._state     = NEW
        self._result    = None
        self._exception = None
        self._traceback = None


    # --------------------------------------------------------------------------
    #
    @classmethod
    def Run(self, call, *args, **kwargs):
        """
        This is a shortcut to 

          f = ru.Future(callable); f.start()

        """

        t = self(call, *args, **kwargs)
        t.start()
        return t


    # --------------------------------------------------------------------------
    #
    def run(self):

        try:
            self._state  = RUNNING
            self._result = self._call(*self._args, **self._kwargs)
            self._state  = DONE

        except Exception as e:
            # NOTE: `state` and `exception` updates are racing.
            tb = traceback.format_exc()
            self._traceback = tb
            self._exception = e
            self._state     = FAILED


    # --------------------------------------------------------------------------
    #
    def wait(self, timeout=None):

        if self.is_alive():
            self.join(timeout=timeout)


    # --------------------------------------------------------------------------
    #
    def cancel(self):

        # FIXME: this is not really implementable generically, so we ignore 
        #        cancel requests for now.  This *should* be handled on the SAGA
        #        layer where cancel requests are forwarded to the callable's
        #        adaptor.
        pass


    # --------------------------------------------------------------------------
    #
    @property
    def state(self):     return self._state 

    @property
    def result(self):    return self._result

    @property
    def exception(self): return self._exception

    @property
    def traceback(self): return self._traceback


# ------------------------------------------------------------------------------





import os
import sys
import math
import shlex

from typing       import Any

from ..which      import which
from ..ids        import generate_id, ID_SIMPLE
from ..modules    import import_module
from ..shell      import sh_callout


# ------------------------------------------------------------------------------
#
class FluxModule(object):

    _flux_core = None
    _flux_job  = None
    _flux_exc  = None
    _flux_v    = None
    _flux_mode = None


    # --------------------------------------------------------------------------
    #
    def __init__(self):
        '''
        import the flux module, if available
        '''

        if self._flux_core or self._flux_job or self._flux_exc:
            return

        flux      = None
        flux_job  = None
        flux_exc  = None
        flux_v    = None
        flux_mode = None

        try:
            out, err, ret = sh_callout('flux --version')
            assert not ret, ['flux --version', err]

            flux_v = out.strip()

            flux     = import_module('flux')
            flux_job = import_module('flux.job')
            if 'JournalConsumer' in dir(flux_job):
                flux_mode = 1
            else:
                flux_mode = 0

        except Exception as e:
            flux_exc = e


        # on failure, try to derive module path from flux executable
        if flux is None or flux_job is None:

            to_pop = None
            try:
                cmd = 'flux python -c "import flux; print(flux.__file__)"'
                out, err, ret = sh_callout(cmd)

                assert not ret, [cmd, err]

                flux_path = os.path.dirname(out.strip())
                mod_path  = os.path.dirname(flux_path)
                sys.path.append(mod_path)
                to_pop = mod_path

                flux     = import_module('flux')
                flux_job = import_module('flux.job')
                if 'JournalConsumer' in dir(flux_job):
                    flux_mode = 1
                else:
                    flux_mode = 0

            except Exception as e:
                flux_exc = e

            if to_pop:
                sys.path.remove(to_pop)

        self._flux_core = flux
        self._flux_job  = flux_job
        self._flux_exc  = flux_exc
        self._flux_v    = flux_v
        self._flux_mode = flux_mode
        self._flux_exe  = which('flux')


    # --------------------------------------------------------------------------
    #
    def verify(self) -> None:
        '''
        verify that flux modules are available
        '''

        if self._flux_core is None:
            raise RuntimeError('flux core module not found') from self._flux_exc

        if self._flux_job is None:
            raise RuntimeError('flux.job module not found') from self._flux_exc

        if self._flux_exe is None:
            raise RuntimeError('flux executable not found') from self._flux_exc


    # --------------------------------------------------------------------------
    #
    @property
    def mode(self) -> int:
        return self._flux_mode

    @property
    def version(self) -> int:
        return self._flux_v

    @property
    def core(self) -> Any:
        return self._flux_core

    @property
    def job(self) -> Any:
        return self._flux_job

    @property
    def exc(self) -> Any:
        return self._flux_exc

    @property
    def exe(self) -> str:
        return self._flux_exe


# ------------------------------------------------------------------------------
#
def spec_from_command(cmd: str) -> 'flux.job.JobspecV1':

    fm = FluxModule()

    spec = fm.job.JobspecV1.from_command(shlex.split(cmd))
    if not 'user' in spec.attributes:
        spec.attributes['user'] = dict()
    spec.attributes['user']['uid'] = generate_id(ID_SIMPLE)

    return spec


# ------------------------------------------------------------------------------
#
def spec_from_dict(td: dict) -> 'flux.job.JobspecV1':

    fm = FluxModule()

    version = 1
    user    = {'uid'     : td.get('uid', generate_id('ru_flux', ID_SIMPLE))}
    system  = {'duration': td.get('timeout', 0.0)}
    tasks   = [{'command': [td['executable']] + td.get('arguments', []),
                'slot'   : 'task',
                'count'  : {'per_slot': 1}}]

    if 'environment' in td: system['environment'] = td['environment']
    if 'sandbox'     in td: system['cwd']         = td['sandbox']
    if 'shell'       in td: system['shell']       = td['shell']

    attributes = {'system' : system,
                  'user'   : user}

    n_ranks = td.get('ranks', 1)
    resources  = [{'count': n_ranks,
                   'type' : 'slot',
                   'label': 'task',
                   'with' : [{
                       'count': int(td.get('cores_per_rank', 1)),
                       'type' : 'core'}]}]
                 #     'count': int(td.get('gpus_per_rank', 0)) or None,
                 #     'type' : 'gpu'

    gpr = td.get('gpus_per_rank', 0)
    if gpr:
        resources[0]['with'].append({'count': math.ceil(gpr),  # flux needs int
                                     'type' : 'gpu'})

    spec = fm.job.JobspecV1(resources=resources,
                            attributes=attributes,
                            tasks=tasks,
                            version=version)

    if n_ranks > 1:
        if td.get('use_mpi', True):
            # ensure that all ranks exit if one rank fails
            spec.setattr_shell_option('exit-on-error', 1)  # defaults to 0
        else:
            spec.setattr_shell_option('exit-timeout', 'none') # defaults to 30

    if td.get('stdin') : spec.stdin  = td['stdin']
    if td.get('stdout'): spec.stdout = td['stdout']
    if td.get('stderr'): spec.stderr = td['stderr']

    return spec


# ------------------------------------------------------------------------------





__author__    = "Radical.Utils Development Team"
__copyright__ = "Copyright 2013, RADICAL@Rutgers"
__license__   = "MIT"

import math

from .scheduler_base import SchedulerBase


# ------------------------------------------------------------------------------
#
class BitarrayScheduler(SchedulerBase):

    try:
        from bitarray import bitarray as _ba             # pylint: disable=E0401

    except:
        # fake with a do-nothing implementation so that initialization works
        class _ba(object):
            def __init__(self, i):
                pass
            def setall(self, x):
                pass

    _one = _ba(1)
    _one.setall(True)

    # --------------------------------------------------------------------------
    #
    def __init__(self, resources=0):

        if 'cores' not in resources:
            raise ValueError('no cores to schedule over')

        self._pos  = 0
        self._size = resources.get('cores')
        self._ppn  = resources.get('ppn', self._size)

        self._cores = self._ba(self._size)
        self._cores.setall(True)   # True: free

        self._flag_align   = resources.get('align',   True)
        self._flag_scatter = resources.get('scatter', True)


    # --------------------------------------------------------------------------
    #
    def get_layout(self):

        return {'cores' : self._size,
                'rows'  : math.sqrt(self._size),
                'cols'  : math.sqrt(self._size)}


    # --------------------------------------------------------------------------
    #
    def get_map(self):

        return self._cores


    # --------------------------------------------------------------------------
    #
    def _align(self, pat, req, loc):
        """
        For a given continuous set of cores, try to move the set so that it
        resides within a single node.  This obviously only works for sets
        smaller than self.ppn.
        """

        if not self._flag_align:
            # alignment is disabled
            return loc, False

        if req > self._ppn:
            # we can't align allocations spanning nodes
            return loc, False

        # we check if the allocation is on a single node.  If that is not the
        # case, we search again from the point of allocation, and check again.
        # We do that until we cycle around the full core list once, reaching
        # the original location again
        orig_pos   = loc
        pos        = loc
        start_node =  pos            / self._ppn
        end_node   = (pos + req - 1) / self._ppn

        if start_node == end_node:
            # already aligned
            return loc, False

        # we search from pos=loc to the end of cores, then rewind to pos=0
        # and search again from the beginning until again pos=loc.
        rewound = False
        while (pos <= self._size and not rewound) or \
              (pos <  orig_pos   and     rewound)    :

            if pos > self._size:
                pos     = 0
                rewound = True

            # we need to renew the allocation -- we search further from pos
            # until we find something
            loc = self._cores.search(pat, 1, pos + 1)

            if not loc:
                pos     =  0
                rewound = True
                continue

            pos        = loc[0]
            start_node =  pos            / self._ppn
            end_node   = (pos + req - 1) / self._ppn

            if start_node == end_node:
                # found an alignment
                return pos, True

            # found new loc, but no dice wrt. alignment -- try again


        raise RuntimeError('alignment error (%s cores requested)' % req)


    # ------------------------------------------------------------------------------
    #
    def alloc(self, req):

        scattered = False
        aligned   = False
        orig_pos  = self._pos
        loc       = list()
        pat       = self._ba(req)   # keep dict of patterns?
        pat.setall(True)

        # simple search
        loc = self._cores.search(pat, 1, self._pos)


        if not loc:
            # FIXME: assume we first search from pos 100 to 1000, then rewind,
            #        then search from 0 to 1000, 900 cores being searched twice.
            #        We should add an 'end_pos' parameter to bitarray.search.
            #        But then again, rewind should be rare if cores are on
            #        average much larger than requests -- and if that is not the
            #        case, we won't be able to allocate many requests anyway...
            self._pos = 0
            loc = self._cores.search(pat, 1, self._pos)


        if not loc and self._flag_scatter:

            scattered = True

            # search for non-continuous free cores
            loc = self._cores.search(self._one, req, orig_pos)

            if not loc:
                # try again from start
                loc = self._cores.search(self._one, req, 0)


        if not loc:
          # with ru_open('ba_cores.bin', 'w') as f:
          #     self._cores.tofile(f)
          # with ru_open('ba_pat.bin', 'w') as f:
          #     pat.tofile(f)
          # with ru_open('ba_req.bin', 'w') as f:
          #     f.write('%d\n' % req)
            self._pos = 0
            raise RuntimeError('out of cores (%s cores requested)' % req)


        # found a match - update pos
        if scattered:

            # we got a scattered list of cores
            self._pos = loc[-1] + 1
            self._cores.setlist(loc, False)
          # for i in loc:
          #     # FIXME: bulk op?
          #     self._cores[i] = False

        else:
            # we got a continuous block
            loc = loc[0]

            if self._flag_align:
                loc, aligned = self._align(pat, req, loc)

            self._pos = loc + req
            self._cores.setrange(loc, self._pos - 1, False)


        return [req, loc, scattered, aligned]


    # --------------------------------------------------------------------------
    #
    def dealloc(self, res):

        req, loc, scattered, _ = res

        if scattered:
            # we got a scattered list of cores
            self._cores.setlist(loc, True)
          # for i in loc:
          #     # FIXME: bulk op?
          #     self._cores[i] = True
        else:
            # we got a continuous block
            self._cores.setrange(loc, loc + req - 1, True)


    # --------------------------------------------------------------------------
    #
    def get_stats(self):

        # determine frequency and length of stretches of free and busy cores
        free_dict = dict()
        free_cnt  = 0
        busy_dict = dict()
        busy_cnt  = 0
        for c in self._cores:
            if c:
                free_cnt += 1
                if busy_cnt:
                    if busy_cnt not in busy_dict:
                        busy_dict[busy_cnt] = 0
                    busy_dict[busy_cnt] += 1
                    busy_cnt = 0
            else:
                busy_cnt += 1
                if free_cnt:
                    if free_cnt not in free_dict:
                        free_dict[free_cnt] = 0
                    free_dict[free_cnt] += 1
                    free_cnt = 0

        # determine frequency of free core counts per node
        node = 0
        node_free = dict()
        for i in range(self._ppn + 1):
            node_free[i] = 0
        while node * (self._ppn + 1) < self._size:
            free_count = 0
            for i in range(self._ppn):
                if self._cores[node * self._ppn + i]:
                    free_count += 1
            node_free[free_count] += 1
            node += 1


        return {'total'     : self._size,
                'free'      : self._cores.count(),
                'busy'      : self._size - self._cores.count(),
                'free_dist' : free_dict,
                'busy_dist' : busy_dict,
                'node_free' : node_free}


# ------------------------------------------------------------------------------





__author__    = 'Radical.Utils Development Team (Andre Merzky)'
__copyright__ = 'Copyright 2013, RADICAL@Rutgers'
__license__   = 'MIT'

import re
import fnmatch

PRESERVE   = 'preserve'
OVERWRITE  = 'overwrite'


# see http://code.activestate.com/recipes/117236-dictionary-mixin-framework/

# ------------------------------------------------------------------------------
#
class DictMixin(dict):
    '''
    Mixin defining all dictionary methods for classes that already have
    a minimum dictionary interface including getitem, setitem, delitem,
    and keys.  Based on those methods, the mixin provides the remaining
    interface functionality to make the class look like a fully compliant
    dictionary.
    '''

  # # --------------------------------------------------------------------------
  # #
  # # mascerade as dict for `isinstance` calls
  # #
  # @property
  # def __class__(self):
  #     return dict


    # --------------------------------------------------------------------------
    #
    # first level definitions should be implemented by the sub-class
    #
    def __getitem__(self, key):
        raise NotImplementedError

    def __setitem__(self, key, value):
        raise NotImplementedError

    def __delitem__(self, key):
        raise NotImplementedError

    def keys(self):
        raise NotImplementedError


    # --------------------------------------------------------------------------
    #
    # second level definitions which assume only getitem and keys
    #
    def has_key(self, key):
        return key in list(self.keys())

    def __iter__(self):
        for k in list(self.keys()):
            yield k


    # --------------------------------------------------------------------------
    #
    # third level uses second level instead of first
    #
    def __contains__(self, key):
        return self.has_key(key)                                    # noqa: W601

    def items(self):
        for k in self:
            yield k, self[k]


    # --------------------------------------------------------------------------
    #
    # fourth level uses second and third levels instead of first
    #
    def iterkeys(self):
        return self.__iter__()

    def itervalues(self):
        for _, v in self.items():
            yield v

    def clear(self):
        for key in list(self.keys()):
            del self[key]

    def setdefault(self, key, default):
        if key not in self:
            self[key] = default
            return default
        return self[key]

    def popitem(self):
        key = list(self.keys())[0]
        value = self[key]
        del self[key]
        return key, value

    def update(self, other):
        for key in list(other.keys()):
            self[key] = other[key]

    def get(self, key, default=None):
        if key in self:
            return self[key]
        return default

    def __repr__(self):
        return repr(dict(list(self.items())))

    def __len__(self):
        return len(self.keys())


# ------------------------------------------------------------------------------
#
def dict_merge(a, b, policy=None, wildcards=False, log=None, _path=None):
    # thanks to
    # http://stackoverflow.com/questions/7204805/ \
    #                          python-dictionaries-of-dictionaries-merge
    '''
    This merges two dict in place, modifying the original dict in a.

    Merge Policies:
      - None (default) : raise an exception on conflicts
      - OVERWRITE      : values in a are overwritten by new values from b
      - PRESERVE       : original value in a are preserved, new values
                         from b are only added where the original value
                         is not set.

    '''

    if  a    is None: return a
    if  b    is None: return a
    if _path is None: _path = list()

    if  not isinstance(a, dict):
        raise TypeError('*dict*_merge expects dicts, not %s' % type(a))

    if  not isinstance(b, dict):
        raise TypeError('*dict*_merge expects dicts, not %s' % type(b))

    # --------------------------------------------------------------------------
    def merge_key(a, key_a, b, key_b):

        # need to resolve conflict
        if isinstance(a[key_a], dict) and isinstance(b[key_b], dict):
            dict_merge(a[key_a], b[key_b],
                       policy    = policy,
                       wildcards = wildcards,
                       log       = log,
                       _path     = _path + [str(key_a)])

        elif key_a not in a and key_b in b:
            a[key_a] = b[key_b]  # use b value

        elif key_a in a and key_b not in b:
            pass  # keep a value

        elif a[key_a] == b[key_b]:
            pass  # same leaf value

        elif key_a not in a and key_b not in b:
            pass  # keep no a value

        else:
            if  policy == PRESERVE:
                if  log:
                    log.debug('preserving key %s:%s \t(%s)'
                              % (':'.join(_path), key_b, b[key_b]))

            elif policy == OVERWRITE:
                if  log:
                    log.debug('overwriting key %s:%s \t(%s)'
                              % (':'.join(_path), key_b, b[key_b]))
                a[key_a] = b[key_b]  # use new value
            else:
                raise ValueError('Conflict at %s (%s : %s)'
                              % ('.'.join(_path + [str(key_a)]),
                                 a[key_a], b[key_b]))
    # --------------------------------------------------------------------------

    # first a clean merge, i.e. no interpretation of wildcards
    for key in sorted(b.keys()):

        if  key in a:

            # need to resolve conflict
            merge_key(a, key, b, key)

        else:
            # no conflict - simply add.  Not that this is a potential shallow
            # copy if b[key] is a complex type.
            a[key] = b[key]


    # optionally, check if other merge options are also valid
    for key_b in sorted(b.keys()):
        if  wildcards:
            if  '*' in key_b:
                pat = re.compile(fnmatch.translate(key_b))
                for key_a in sorted(a.keys()):
                    if  pat.match(key_a):
                        merge_key(a, key_a, b, key_b)

    return a


# ------------------------------------------------------------------------------
#
def dict_stringexpand(target, sources=None):
    '''
    This expands dict entries (strings only) with keys from a second dict. For
    example, the dicts::

        target  = {'workdir'  : '/home/%(user)s/',
                   'resource' : '%(resource)s'}
        sources = {'user'     : 'peer_gynt',
                   'protocol' : 'ssh',
                   'host'     : 'localhost',
                   'resource' : '%(protocol)s://%(host)s/'}

    would result in::
        target = {'workdir'  : '/home/peer_gynt/',
                  'resource' : 'ssh://localhost'}

    Note that expansion happened twice, for the `resource` tag to be fully
    specified.
    '''

    assert isinstance(target, dict)

    # expand from self, and all given dicts, but only use
    # first-level primitive types (string, int, float)
    if  sources:
        if  isinstance(sources, dict):
            sources = [sources]
    else:
        sources = list()

    if  not isinstance(sources, list):
        raise TypeError('Need dict as expansion source, not %s' % type(sources))

    # target must be first source, to avoid cycles (other sources are likely to
    # have *other* info)
    sources.insert(0, target)

    repl_source = dict()
    for source in sources:
        for key, val in source.items():
            if  isinstance(val, str) or \
                isinstance(val, int       ) or \
                isinstance(val, float     ) :
                repl_source[key] = val

    again = True
    while again:
        target, again = _generic_stringexpand(target, repl_source)

    return target


# ------------------------------------------------------------------------------
#
def _generic_stringexpand(target, source):

    if  isinstance(target, str):
        return _string_stringexpand(target, source)

    elif  isinstance(target, list):
        return _list_stringexpand(target, source)

    elif  isinstance(target, dict):
        return _dict_stringexpand(target, source)

    else:
        # ignore other types for now
        return target, False


# ------------------------------------------------------------------------------
#
def _list_stringexpand(target, source):

    assert isinstance(target, list)
    assert isinstance(source, dict)

    all_again = 0
    for idx, elem in enumerate(target):
        target[idx], again = _generic_stringexpand(elem, source)
        all_again += again

    return target, all_again


# ------------------------------------------------------------------------------
#
def _dict_stringexpand(target, source):

    assert (isinstance(target, dict))
    assert (isinstance(source, dict))

    all_again = 0
    for key, val in target.items():
        target[key], again = _generic_stringexpand(val, source)
        all_again += again

    return target, all_again


# ------------------------------------------------------------------------------
#
def _string_stringexpand(target, source):

    assert (isinstance(target, str))
    assert (isinstance(source, dict))

    orig = str(target)
    try:
        expanded = target % source

    except KeyError:
        # ignore incomplete expands
        return orig, False

    except ValueError:
        # ignore incomplete expands
        return orig, False

    # only check for success after success.  Duh!
    if  orig == expanded: return expanded, False
    else                : return expanded, True


# ------------------------------------------------------------------------------
#
def iter_diff(a, b):

    if isinstance(a, list) and isinstance(b, list):
        for ea, eb in zip(a, b):
            iter_diff(ea, eb)
    elif isinstance(a, dict) and isinstance(b, dict):
        assert len(list(a.keys())) == len(list(b.keys())), \
               (len(a), sorted(list(a.keys())), len(b), sorted(list(b.keys())))
        for k in a:
            iter_diff(a[k], b[k])
    else:
        if a != b:
            print('elem diff: %s' % a)
            print('elem diff: %s' % b)
            print()
        assert a == b, [a, b]


# ------------------------------------------------------------------------------
#
def dict_diff(a, b):
    '''
    return a dict of the form::

        {
           'k1': {'a': 'foo',
                  'b': 'bar'},
           'k2': {'a': 'foo'},
           'k3': {'b': 'bar'},
        }

    which contains only those keys which are different in the two given dicts.
    Keys which are missing in either one are not included (to distinguish from
    `None` values).  This methods operates recursively over the given dicts.
    '''

    def _list_diff(a, b):

        if len(a) != len(b):
            la, lb = sorted([len(a), len(b)])
            ret = ['len(%d) != len(%d)' % (la, lb)]
        else:
            ret = list()
            for va, vb in zip(a, b):
                if isinstance(va, dict) and isinstance(vb, dict):
                    tmp = _dict_diff(va, vb)
                    if tmp:
                        ret.append(tmp)
                elif isinstance(va, list) and isinstance(vb, list):
                    tmp = _list_diff(va, vb)
                    if tmp:
                        ret.append(tmp)
                else:
                    if va != vb:
                        ret.append({'a': va,
                                    'b': vb})

        return ret


    def _dict_diff(a, b):

        ka  = sorted(list(a.keys()))
        kb  = sorted(list(b.keys()))
        ret = dict()

        for k in ka:
            if k not in kb:
                ret[k] = {'a': a[k]}
            else:
                va = a[k]
                vb = b[k]
                if isinstance(va, dict) and isinstance(vb, dict):
                    tmp = _dict_diff(va, vb)
                    if tmp:
                        ret[k] = tmp
                elif isinstance(va, list) and isinstance(vb, list):
                    tmp = _list_diff(va, vb)
                    if tmp:
                        ret[k] = tmp
                elif va != vb:
                    ret[k] = {'a': va,
                              'b': vb}

        for k in kb:
            if k not in ka:
                ret[k] = {'b': b[k]}
            else:
                va = a[k]
                vb = b[k]
                if isinstance(vb, dict) and isinstance(va, dict):
                    tmp = _dict_diff(vb, va)
                    if tmp:
                        ret[k] = tmp
                elif isinstance(va, list) and isinstance(vb, list):
                    tmp = _list_diff(va, vb)
                    if tmp:
                        ret[k] = tmp
                elif va != vb:
                    ret[k] = {'a': va,
                              'b': vb}

        return ret

    return _dict_diff(a, b)


# ------------------------------------------------------------------------------





__author__    = "Radical.Utils Development Team (Andre Merzky, Ole Weidner)"
__copyright__ = "Copyright 2013, RADICAL@Rutgers"
__license__   = "MIT"


import os


# ------------------------------------------------------------------------------
#
def which(names):
    '''
    Takes a (list of) name(s) and looks for an executable in the path.  It
    will return the first match found, or `None` if none of the given names
    is found. 
    '''

    def is_exe(fpath):
        return os.path.isfile(fpath) and os.access(fpath, os.X_OK)

    if not isinstance(names, list):
        names = [names]

    for name in names:

        if is_exe(name):
            return name

        for path in os.environ.get('PATH', '').split(':'):
            fpath = '%s/%s' % (path, name)
            if is_exe(fpath):
                return fpath

    return None


# ------------------------------------------------------------------------------





__author__    = "Radical.Utils Development Team (Andre Merzky)"
__copyright__ = "Copyright 2013, RADICAL@Rutgers"
__license__   = "MIT"


import os
import threading

from .atfork import atfork


_singleton_lock = threading.RLock ()


''' Provides a Singleton metaclass.  '''


# ------------------------------------------------------------------------------
#
class Singleton (type):
    '''
    A metaclass to 'tag' other classes as singleton::

        class MyClass(BaseClass):
            __metaclass__ = Singleton

    '''
    _instances = {}

    def __call__(cls, *args, **kwargs):

        with _singleton_lock:

            if  cls not in cls._instances:
                cls._instances[cls] = super().__call__(*args, **kwargs)

            return cls._instances[cls]


# ------------------------------------------------------------------------------
#
def _atfork_prepare():
    pass


def _atfork_parent():
    pass


def _atfork_child():
    # release lock
    global _singleton_lock
    _singleton_lock = threading.RLock()


# lock cleaning can be disabled by setting RADICAL_UTILS_NO_ATFORK
if 'RADICAL_UTILS_NO_ATFORK' not in os.environ:
    atfork(_atfork_prepare, _atfork_parent, _atfork_child)


# ------------------------------------------------------------------------------





'''
Config file parsing.

We provide a json based config file parser with following properties

  - system config files will be merged with user configs (if those exist)
  - python style comments are filtered out before parsing
  - after parsing, `${ABC:default}`-style values are set or expanded via
    `os.environ`
  - the returned class exposes settings as dicts or attributes
    cfg['foo']['bar'] == cfg.foo.bar


Config Names and Locations
--------------------------

We assume two basic locations for config files: one is installed within the
scope of a Python module, the other one is under user control, and usually in
the users home directory.  The  config reader accepts the following parameters
to derive the exact locations:

  - module: name of module under which the config is installed
  - path  : config file path relative to the module home
  - name  : config file name relative to the path

The `module` string is interpreted as follows:

    module         = 'module'
    module_path    = radical.utils.debug.find_module(module)
    sys_config_dir = '%s/configs' % module_path
    usr_base_dir   = os.getenv('RADICAL_CONFIG_USER_DIR') or \
                     os.getenv('HOME')
    if usr_base_dir:
        usr_config_dir = '%s/.%s/configs' % \
            (usr_base_dir, module.replace('.', '/'))

so the location of the module's `__init__.py` is used to derive the location
of the installed system config files, and the module name is used to derive
the location of the user provided config files.

For example, the module `radical.utils` will have the following config dirs:

    sys_config_dir = /tmp/ve/lib/python3.8/site-packages/radical/utils/configs
    usr_config_dir = /home/merzky/.radical/utils/configs

After loading the system level config file, any existing user level config
file is merged into it, via

    radical.utils.dict_merge(user_cfg, system_cfg, mode='overwrite')

so that the user config settings supersede the system config settings.

Both path and name specifiers can contain `*` as wildcard, which is then
interpreted as by `glob()`.  If that wildcard exists, then all matching config
files are read into *one* configuration dict, where each root key is set to
the value the '*' expands to (minus the `.json` extension).

For example, the name `radical.pilot.resource_*` with the following config
files:

    /tmp/ve/[...]/radical/pilot/configs/resource_xsede.json
    /tmp/ve/[...]/radical/pilot/configs/resource_ncsa.json

will result in a config dict like::

    {'xsede' : { 'foo' : 'bar' },
     'ncsa'  : { 'fiz' : 'baz' }}


Queries
-------

We support two types of queries on the resulting parsed configs:

  - `dict` like queries
  - the `_query(key)` method returns a single value, or 'None' if not found.

In the latter `_query()` case, the `key` can be specified as dot-separated
path, so that the following two snippets are equivalent (assuming that a
`foo.bar` section exists):

  val = cfg['foo']['bar'].get('baz')
  val = cfg._query('foo.bar.baz')


Environment
-----------

Towards `os.environ` completion, we support the following syntax in all string
*values* (not keys):

    `${RADICAL_UTILS_ENV:default_value}`

which will be replaced by

    `os.environ.get('RADICAL_UTILS_ENV', 'default_value')`

The default value is optional, an empty string is used if no default value is
given.  Env evaluation is only performed at time of parsing, not at time of
query.  RU attempts to convert env variables to float and int - if that fails,
values are stored as strings.


Validation
----------

It probably makes sense to switch to a json schema validator at some point,
see for example https://pypi.python.org/pypi/json-schema-validator. For now
this implementation remains schema-less, and will thus, in a very pythonesque
way, only fail once values are queried or used.


Implementation
--------------

This implementation is based on typed dictionaries which are accessed as
`TypedDict`'ed object hierarchy.
'''

__author__    = 'RADICAL-Cybertools Team'
__copyright__ = 'Copyright 2016-2022, The RADICAL-Cybertools Team'
__license__   = 'MIT'

import glob
import os

from .modules    import find_module
from .misc       import expand_env as ru_expand_env
from .json_io    import read_json, write_json
from .dict_mixin import dict_merge
from .typeddict  import TypedDict, TypedDictMeta

from .singleton  import Singleton


# ------------------------------------------------------------------------------
#
class Config(TypedDict):
    """Contents of a config (json) file from a module's config tree.

    Attributes:
        module:    used to determine the module's config file location
                   (default: `radical.utils`)
        category:  name of config to be loaded from module's config path
        name:      specify a specific configuration to be used
        path:      path to app config json to be used for initialization
        cfg:       application config dict to be used for initialization
        from_dict: alias for cfg, to satisfy base class constructor
        expand:    enable / disable environment var expansion (default: True)
        env:       environment dictionary to be used for expansion
                   (default: `os.environ`)

    The naming of config files follows this rule:

        `<category>_<name>.json`

    For example, if the following is used in a system python installation:

        ru.Config('radical.pilot', category='session', name='minimal')

    it would attempt to load (depending on system details):

        /usr/lib/python3/site-packages/ \
             radical/pilot/configs/session_minimal.json

    NOTE: Keys containing an underscore are not exposed via the API.
          Keys containing dots are split and interpreted as paths in the
          configuration hierarchy.
    """

    _self_default = True

    # FIXME: we should cache config files after reading, so that repeated
    #        instance creations do not trigger a new (identical) round of
    #        parsing.

    # --------------------------------------------------------------------------
    #
    def __init__(self, from_dict=None,
                       module=None, category=None, name=None,
                       cfg=None,    path=None,     expand=True,
                       env=None, _internal=False):
        """
        Load a config (json) file from the module's config tree, and overload
        any user specific config settings if found.

        Parameters
        ----------
        from_dict: alias for cfg, to satisfy base class constructor
        module:    used to determine the module's config file location
                   - default: `radical.utils`
        category:  name of config to be loaded from module's config path
        name:      specify a specific configuration to be used
        path:      path to app config json to be used for initialization
        cfg:       application config dict to be used for initialization
        expand:    enable / disable environment var expansion
                   - default: True
        env:       environment dictionary to be used for expansion
                   - default: `os.environ`

        The naming of config files follows this rule:

          `<category>_<name>.json`

        For example, if the following is used in a system python installation:

            ru.Config('radical.pilot', category='session', name='minimal')

        it would attempt to load (depending on system details):

            /usr/lib/python3/site-packages/\
                 radical/pilot/configs/session_minimal.json

        NOTE: Keys containing an underscore are not exposed via the API.
              Keys containing dots are split and interpreted as paths in the
              configuration hierarchy.
        """

        # if the `from_dict` is given but is a string, we interpret it as
        # `module` parameter.
        if from_dict and isinstance(from_dict, str):
            module    = from_dict
            from_dict = None

        if from_dict:
            # if we could only overload constructors by signature... :-/
            # As it is, we have to emulate that...
            assert (not module   and
                    not category and
                    not name     and
                    not cfg      and
                    not path     and
                    not env      and
                    not _internal), 'from_dict must be exclusive'
            super().__init__(from_dict=from_dict)
            return

        if path and cfg:
            raise ValueError('conflicting initializers (path, cfg)')

        if path:
            cfg = read_json(path)

        if not cfg:
            # just use config files
            cfg = dict()

        # if a category has dot limited elements and no module is given,
        # interpret the first part as module
        # radical.pilot.session -> [radical.pilot, session]
        if not module and category and '.' in category:
            elems    = category.split('.')
            module   = '.'.join(elems[:-1])
            category = elems[-1]

        # do the same if module is given but not the category
        if not category and module and '.' in module:
            elems    = module.split('.')
            module   = '.'.join(elems[:-1])
            category = elems[-1]

        # if a category starts with a module prefix, strip that prefix
        if category and category.startswith('%s.' % module):
            category = category[len(module) + 1:]

        name_orig = name
        if not name:
            # by default, load the default config
            name = 'default'

        # purge module and category parts from name
        if module and name.startswith(module + '.'):
            name = name[len(module) + 1:]

        if category and name.startswith(category + '.'):
            name = name[len(category) + 1:]

        if name.startswith('/'):

            # load config from abs path
            sys_fspec = name
            usr_fspec = None

            if '*' in name: starred = True
            else          : starred = False

        elif module and category:

            # construct cfg file paths
            modpath = find_module(module)
            if not modpath:
                raise ValueError("Cannot find module %s" % module)

            fname     = '%s_%s.json' % (category.replace('.', '/'), name)

            sys_dir   = '%s/configs' % modpath
            sys_fspec = '%s/%s'      % (sys_dir, fname)

            home = os.getenv('RADICAL_CONFIG_USER_DIR') or os.getenv('HOME')
            if home:
                usr_dir   = '%s/.%s/configs' % (home, module.replace('.', '/'))
                usr_fspec = '%s/%s'          % (usr_dir, fname)
            else:
                usr_dir   = None
                usr_fspec = None

            if '*' in fname: starred = True
            else           : starred = False

        else:
            # we can't load a config file - just use the app config
            sys_fspec = None
            usr_fspec = None
            starred   = False

        sys_cfg = dict()
        usr_cfg = dict()
        app_cfg = cfg

        if _internal:
            # no need to look at the FS, just convert the given cfg dict
            sys_fspec = None
            usr_fspec = None

        nfiles = 0
        if not starred:

            if sys_fspec:

                sys_fname = sys_fspec

                if not os.path.isfile(sys_fname):
                    sys_fname += '.json'

                if os.path.isfile(sys_fname):
                    sys_cfg = read_json(sys_fname)
                    nfiles += 1

            if usr_fspec:

                usr_fname = usr_fspec

                if not os.path.isfile(usr_fname):
                    usr_fname += '.json'

                if os.path.isfile(usr_fname):
                    usr_cfg = read_json(usr_fname)
                    nfiles += 1

        else:

            # wildcard mode: whatever the '*' expands into is used as root dict
            # entry, and the respective content of the config file is stored
            # underneath it.
            if sys_fspec:

                postfix_len = len('.json')                      # ' .json'
                prefix_len  = len(sys_fspec) - postfix_len - 1  # '*.json'

                for sys_fname in glob.glob(sys_fspec):

                    base = sys_fname[prefix_len:-postfix_len]
                    scfg = read_json(sys_fname)
                    sys_cfg[base] = scfg
                    nfiles += 1

            if usr_fspec:

                postfix_len = len('.json')                      # ' .json'
                prefix_len  = len(usr_fspec) - postfix_len - 1  # '*.json'

                for usr_fname in glob.glob(usr_fspec):

                    base = usr_fname[prefix_len:-postfix_len]
                    ucfg = read_json(usr_fname)
                    usr_cfg[base] = ucfg
                    nfiles += 1

        # if we did not find *any* file, and the original `name` was None,
        # then try to load config files w/o name
        # Example: if there is no `registry_default.json`, then try to load
        # `registry.json`.
        if nfiles == 0 and name_orig is None and not _internal and category:

            fname     = '%s.json' % (category.replace('.', '/'))
            sys_fname = '%s/%s'   % (sys_dir, fname)
            if os.path.isfile(sys_fname):
                sys_cfg = read_json(sys_fname)

            if usr_dir:
                usr_fname = '%s/%s' % (usr_dir, fname)
                if os.path.isfile(usr_fname):
                    usr_cfg = read_json(usr_fname)


        # merge sys, usr and app cfg before expansion
        cfg_dict = dict()
        cfg_dict = dict_merge(cfg_dict, sys_cfg, policy='overwrite')
        cfg_dict = dict_merge(cfg_dict, usr_cfg, policy='overwrite')
        cfg_dict = dict_merge(cfg_dict, app_cfg, policy='overwrite')

        if expand:
            cfg_dict = ru_expand_env(cfg_dict, env=env)

        super().__init__(from_dict=cfg_dict)

    # --------------------------------------------------------------------------
    #
    def write(self, fname):

        write_json(self.as_dict(), fname)


# ------------------------------------------------------------------------------
#
class DefaultConfigMeta(TypedDictMeta, Singleton):
    """
    Metaclass inherited from `TypedDict`'s metaclass along with `Singleton` to
    avoid metaclass conflict (the metaclass of a derived class must be a
    subclass of the metaclasses of all its bases). `Singleton` metaclass allows
    to have only one instance of a corresponding class.
    """
    pass


# ------------------------------------------------------------------------------
#
class DefaultConfig(Config, metaclass=DefaultConfigMeta):
    """
    The settings in this default config are, unsurprisingly, used as default
    values for various RU classes and methods, as for example for log file
    locations, log levels, profile locations, etc.
    """

    _schema = {
        'log_lvl'    : str,
        'log_tgt'    : str,
        'log_dir'    : str,
        'report'     : bool,
        'report_tgt' : str,
        'report_dir' : str,
        'profile'    : bool,
        'profile_dir': str,
        'iface'      : list,
    }

    # --------------------------------------------------------------------------
    #
    def __init__(self):

        super().__init__(module='radical.utils',
                         category='utils',
                         name='default')


# ------------------------------------------------------------------------------





__author__    = "Radical.Utils Development Team"
__copyright__ = "Copyright 2014-2020, RADICAL@Rutgers"
__license__   = "MIT"

# pylint: disable=protected-access

import time
import threading as mt

from   datetime import datetime


# ------------------------------------------------------------------------------
#
# static datetime instance used for the `epoch` helper method
_epoch = datetime(1970, 1, 1)


# ------------------------------------------------------------------------------
#
def epoch(data, pattern):
    '''
    convert a given datetime string into seconds since EPOCH.
    The string is parsed as defined by POSIX's `strptime()`.
    '''

    return dt_epoch(datetime.strptime(data, pattern))


# ------------------------------------------------------------------------------
#
def dt_epoch(dt):
    '''
    convert a given datetime instance into seconds since EPOCH.
    '''

    return (dt - _epoch).total_seconds()


# ------------------------------------------------------------------------------
#
# Provides class decorators to time all public class methods
#
# The class decorator is provided as decorator method `@timed_class`, and as
# decorator class `@TimedClass`.  It thus also serves as documentation on the
# different ways to implement class decorators.
#
# This is called on *decorator class* instantiation
def timed_method(func):
    '''
    This class decorator will decorate all public class methods with a timing
    function.  That will time the call invocation, and store the respective data
    in a private class dict '__timing'.  Additionally, it will add the class
    method '_timing_last()', which will return the tuple `[method name, method
    timer]` for the last timed invocation of a class method.
    '''

    # apply timing decorator to all public methods
    def func_timer(obj, *args, **kwargs):

        try:
            _ = obj.__timing
            assert _

        except Exception:

            # never seen this one before -- create private timing dict, and add
            # timing_last method
            obj.__timing = dict()

            def timing_last():
                last_call = obj.__timing.get('__last_call', None)
                timer     = obj.__timing.get(last_call, [None])[0]
                return last_call, timer

            def timing_stats():
                import math
                ret = ""
                for key in sorted(obj.__timing.keys()):

                    if  key == '__last_call':
                        continue

                    tlist = obj.__timing[key]
                    tnum  = len(tlist)
                    tmean = sum(tlist) / tnum
                    tdev  = [x - tmean for x in tlist]
                    tdev2 = [x * x for x in tdev]
                    tstd  = math.sqrt(sum(tdev2) / tnum)

                    ret  += "%-20s: %10.3fs (+/- %5.3fs) [n=%5d]\n" \
                          % (key, tmean, tstd, tnum)

                return ret


            def timing_reset():
                obj.__timing = dict()

            obj._timing_last  = timing_last
            obj._timing_stats = timing_stats
            obj._timing_reset = timing_reset


        # object is set up -- time the call and record timings
        fname = func.__name__
        tdict = obj.__timing

        start = time.time()
        ret   = func(obj, *args, **kwargs)
        stop  = time.time()
        timed = stop - start

        if fname not in tdict:
            tdict[fname] = list()
        tdict[fname].append(timed)
        tdict['__last_call'] = fname

        return ret

    return func_timer


# ------------------------------------------------------------------------------
#
class Time(object):
    '''
    This is a timing class that allows to simulate different types of clocks.

    Parameters:
    tick: This is the resolution of the clock.
    speed: This is the speed of the clock.
    '''


    # --------------------------------------------------------------------------
    #
    def __init__(self, tick=0.01, speed=1.0):
        self._tick    = tick
        self._speed   = speed
        self._seconds = 0.0
        self._term    = mt.Event()
        self._lock    = mt.Lock()
        self._clock   = mt.Thread(target=self._clock_thread)

        self._clock.daemon = True
        self._clock.start()


    # --------------------------------------------------------------------------
    #
    def stop(self):
        '''
        Stops the clock
        '''
        self._term.set()
        self._clock.join()


    # --------------------------------------------------------------------------
    #
    def _clock_thread(self):

        last = time.time()

        while not self._term.is_set():

            time.sleep(self._tick)
            now = time.time()

            with self._lock:
                self._seconds += (now - last) * self._speed

            last = now


    # --------------------------------------------------------------------------
    #
    def time(self):
        '''
        Returns the current value of the clock
        '''

        return self._seconds


    # --------------------------------------------------------------------------
    #
    def sleep(self, amount):
        '''
        Sleeps for a specific amount of time units. Actual time spent is equal
        to amount divided by speed.
        '''

        time.sleep(amount / self._speed)


    # --------------------------------------------------------------------------
    #
    def advance(self, amount):
        '''
        Advance the clock with a specific amount of time units
        '''
        with self._lock:
            self._seconds += amount


# ------------------------------------------------------------------------------





import builtins
import inspect
import os
import importlib

from typing import Any, Union, Optional

_id_cnt = 0


# ------------------------------------------------------------------------------
#
# to keep RU 2.6 compatible, we provide import_module which works around some
# quirks of __import__ when being used with dotted names. This is what the
# python docs recommend to use.  This basically steps down the module path and
# loads the respective submodule until arriving at the target.
#
def import_module(name):

    mod = __import__(name)
    for s in name.split('.')[1:]:
        mod = getattr(mod, s)
    return mod


# ------------------------------------------------------------------------------
#
# as import_module, but without the import part :-P
#
def find_module(name):

    package = importlib.util.find_spec(name)

    if not package:
        return None

    if '_NamespaceLoader' in str(package.loader):
        # since Python 3.5, loaders differ between modules and namespaces
        return package._path._path[0]                    # pylint: disable=W0212
    else:
        return os.path.dirname(package.loader.get_filename())


# ------------------------------------------------------------------------------
#
# a helper to load functions and classes from user provided source file which
# are *not* installed as modules.  All symbols from that file are loaded, and
# returned is a dictionary with the following structure:
#
# symbols = {'classes'  : {'Foo': <class 'mod_0001.Foo'>,
#                          'Bar': <class 'mod_0001.Bar'>,
#                          ...
#                         },
#            'functions': {'foo': <function foo at 0x7f532d241d40>,
#                          'bar': <function bar at 0x7f532d241d40>,
#                          ...
#                         }
#           }
#
def import_file(path):

    global _id_cnt
    _id_cnt += 1

    uid  = 'mod_%d' % _id_cnt
    spec = importlib.util.spec_from_file_location(uid, path)
    mod  = importlib.util.module_from_spec(spec)

    spec.loader.exec_module(mod)

    symbols = {'functions': dict(),
               'classes'  : dict()}

    for k,v in mod.__dict__.items():
        if not k.startswith('__'):
            if inspect.isclass(v):    symbols['classes'  ][k] = v
            if inspect.isfunction(v): symbols['functions'][k] = v

    return symbols


# ------------------------------------------------------------------------------
#
def get_type(type_name: str) -> Optional[type]:
    '''
    get a type object from a type name (str)
    '''

    # check builtin types
    ret = getattr(builtins, type_name, None)
    if isinstance(ret, type):
        return ret

    # check global types
    ret = globals().get(type_name)
    if isinstance(ret, type):
        return ret

    # check local types of the calling frame
    ret = inspect.currentframe().f_back.f_locals.get(type_name)
    if isinstance(ret, type):
        return ret


# ------------------------------------------------------------------------------
#
def load_class(fpath: str,
               cname: str,
               ctype: Optional[Union[type,str]] = None) -> Optional[Any]:
    '''
    load class `cname` from a source file at location `fpath`
    and return it (the class, not an instance).
    '''

    if not os.path.isfile(fpath):
        raise ValueError('no source file at [%s]' % fpath)

    pname  = os.path.splitext(os.path.basename(fpath))[0]
    spec   = importlib.util.spec_from_file_location(pname, fpath)
    plugin = importlib.util.module_from_spec(spec)

    spec.loader.exec_module(plugin)

    ret = getattr(plugin, cname)

    if ctype:

        if isinstance(ctype, str):
            ctype_name = ctype
            ctype = get_type(ctype_name)

            if not ctype:
                raise ValueError('cannot type check %s' % ctype_name)

        if not issubclass(ret, ctype):
            return None

    return ret


# ------------------------------------------------------------------------------




# Copyright 2009 Google Inc.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# Licensed to the PSF under a Contributor Agreement.
#
# Author: Gregory P. Smith <greg@krypto.org>

# flake8: noqa: F401

from .atfork       import monkeypatch_os_fork_functions, atfork
from .stdlib_fixer import fix_logging_module

# Python 3.13+ has a fix for the logging locks
import sys as _sys
_py_version = float("%d.%d" % _sys.version_info[:2])
if _py_version < 3.13:
     fix_logging_module()

monkeypatch_os_fork_functions()





import os
import sys
import time
import pprint
import signal
import random
import inspect
import traceback

import threading as mt

from .ids     import generate_id
from .misc    import ru_open
from .threads import get_thread_name


# ------------------------------------------------------------------------------
#
_ps_cmd = 'ps -efw'
if sys.platform != 'darwin':
    _ps_cmd += ' --forest'


# ------------------------------------------------------------------------------
#
def get_trace():

    trace = sys.exc_info()[2]

    if trace:
        stack          = traceback.extract_tb(trace)
        traceback_list = traceback.format_list(stack)
        return ''.join(traceback_list)

    else:
        stack          = traceback.extract_stack()
        traceback_list = traceback.format_list(stack)
        return ''.join(traceback_list[:-1])


# ------------------------------------------------------------------------------
#
# pylint: disable=unused-argument
def print_stacktraces(signum=None, sigframe=None):
    '''
    signum, sigframe exist to satisfy signal handler signature requirements
    '''

    this_tid = mt.currentThread().ident

    # if multiple processes (ie. a process group) get the signal, then all
    # traces are mixed together.  Thus we waid 'pid%100' milliseconds, in
    # the hope that this will stagger the prints.
    pid = int(os.getpid())
    time.sleep((pid % 100) / 1000)

    out  = '=========================================================\n'
    out += 'RADICAL Utils -- Debug Helper -- Stacktraces\n'
    out += os.popen("%s | grep ' %s ' | grep -v grep" % (_ps_cmd, pid)).read()

    if _debug_helper:
        out += '---------------------------------------------------------\n'
        if _debug_helper.locks:
            out += 'Locks:\n'
        for name, lock in _debug_helper.locks.items():
            owner = lock.owner
            waits = lock.waits
            if not owner: owner = '-'
            out += '  %-60s: %s %s\n' % (name, owner, waits)

        if _debug_helper.rlocks:
            out += 'RLocks:\n'
        for name, rlock in _debug_helper.rlocks.items():
            owner = rlock.owner
            waits = rlock.waits
            if not owner: owner = '-'
            out += '  %-60s: %s %s\n' % (name, owner, waits)
        out += '---------------------------------------------------------\n'

    try:
        info = get_stacktraces()

    except Exception as e:
        out += 'skipping frame (%s)' % e
        info = None

    if info:

        for tid,tname in info:

            if tid == this_tid: marker = '[active]'
            else              : marker = ''

            out += '---------------------------------------------------------\n'
            out += 'Thread: %s %s\n' % (tname, marker)
            out += '  PID : %s \n'   % os.getpid()
            out += '  TID : %s \n'   % tid

            for fname,line,func,code in info[tid,tname]:

                if code: code = code.strip()
                else   : code = '<no code>'

              # # [:-1]: .py vs. .pyc:/
              # if not (__file__[:-1] in fname and \
              #         func in ['get_stacktraces', 'print_stacktraces']):
                if func not in ['get_stacktraces', 'print_stacktraces']:
                    out += '  File: %s, line %d, in %s\n' % (fname, line, func)
                    out += '        %s\n' % code

    out += '========================================================='

    sys.stdout.write('%s\n' % out)

    if 'RADICAL_DEBUG' in os.environ:
        with ru_open('/tmp/ru.stacktrace.%s.log' % pid, 'w') as f:
            f.write('%s\n' % out)

    sys.stdout.flush()


# ------------------------------------------------------------------------------
#
def get_stacktraces():

    id2name = dict()
    for th in mt.enumerate():
        id2name[th.ident] = th.name

    ret = dict()
    stacklist = list(sys._current_frames().items())      # pylint: disable=W0212
    for tid, stack in stacklist:

        name = id2name.get(tid, 'noname')
        ret[tid, name] = traceback.extract_stack(stack)

    return ret


# ------------------------------------------------------------------------------
#
def print_stacktrace(msg=None, _stack=None):

    if not msg:
        msg = ''

    tname = mt.currentThread().name
    pid   = os.getpid()

    out   = '--------------\n'
    out  += 'RADICAL Utils -- Stacktrace [%s] [%s]\n' % (pid, tname)
    out  += '%s\n' % msg
    out  += os.popen("%s | grep ' %s ' | grep -v grep" % (_ps_cmd, pid)).read()

    if not _stack:
        _stack = get_stacktrace()

    for line in _stack:
        out += line.strip()
        out += '\n'

    out += '--------------\n'

    sys.stdout.write(out)


# ------------------------------------------------------------------------------
#
def get_exception_trace(msg=None):

    return traceback.format_exc().split('\n')


# ------------------------------------------------------------------------------
#
def print_exception_trace(msg=None, exc=None):

    if not exc:
        print_stacktrace(msg=msg, _stack=get_exception_trace())
    else:
        print_stacktrace(msg=msg,
                         _stack=traceback.extract_tb(exc.__traceback__))


# ------------------------------------------------------------------------------
#
def get_stacktrace():

    return traceback.format_stack()[:-1]


# ------------------------------------------------------------------------------
#
def get_caller_name(skip=2):
    '''
    Get a name of a caller in the format module.class.method

    `skip` specifies how many levels of stack to skip while getting caller
    name. skip=1 means 'who calls me', skip=2 'who calls my caller' etc.

    An empty string is returned if skipped levels exceed stack height

    Kudos: http://stackoverflow.com/questions/2654113/ \
            python-how-to-get-the-callers-method-name-in-the-called-method
    '''

    stack = inspect.stack()
    start = 0 + skip

    if len(stack) < start + 1:
        return ''

    pframe = stack[start][0]
    name   = list()
    module = inspect.getmodule(pframe)

    # `modname` can be None when frame is executed directly in console
    # TODO(techtonik): consider using __main__
    if module:
        name.append(module.__name__)

    # detect classname
    if 'self' in pframe.f_locals:
        name.append(pframe.f_locals['self'].__class__.__name__)

    codename = pframe.f_code.co_name

    if codename != '<module>':  # top level usually
        name.append(codename)   # function or a method

    del pframe

    return '.'.join(name)


# ------------------------------------------------------------------------------
#
_verb  = False
if 'RADICAL_DEBUG_VERBOSE' in os.environ:
    _verb = True

_raise_on_state = dict()
_raise_on_lock  = mt.Lock()


# ------------------------------------------------------------------------------
#
def raise_on(tag, log=None, msg=None):
    '''
    The purpose of this method is to artificially trigger error conditions for
    testing purposes, for example when handling the n'th unit, getting the n'th
    heartbeat signal, etc.

    The tag parameter is interpreted as follows: on the `n`'th invocation of
    this method with any given `tag`, an exception is raised, and the counter
    for that tag is reset.

    The limit `n` is set via an environment variable `RU_RAISE_ON_<tag>`, with
    `tag` in upper casing.  The environment will only be inspected during the
    first invocation of the method with any given tag.  The tag counter is
    process-local, but is shared amongst threads of that process.
    '''

    with _raise_on_lock:

        if tag not in _raise_on_state:

            env = os.environ.get('RU_RAISE_ON_%s' % tag.upper())

            if env and env.startswith('RANDOM_'):
                # env is rnd spec
                rate  = int(env[7:])
                limit = 1

            elif env:
                # env is int
                rate  = 1
                limit = int(env)

            else:
                # no env set
                rate  = 1
                limit = 0

            _raise_on_state[tag] = {'count': 0,
                                    'rate' : rate,
                                    'limit': limit}

        _raise_on_state[tag]['count'] += 1

        count = _raise_on_state[tag]['count']
        limit = _raise_on_state[tag]['limit']
        rate  = _raise_on_state[tag]['rate']

        if msg    : info = '%s [%2d / %2d] [%s]' % (tag, count, limit, msg)
        elif _verb: info = '%s [%2d / %2d]'      % (tag, count, limit     )

        if log    : log.debug('raise_on checked   %s' , info)
        elif _verb: print('raise_on checked   %s' % info)

        if limit and count == limit:

            _raise_on_state[tag]['count'] = 0

            if rate == 1:
                val = limit

            else:
                val = random.randint(0, 100)

                if val > rate:
                    if log: log.warning('raise_on ignored   %s [%2d / %2d]',
                                          tag, val, rate)
                    elif _verb:   print('raise_on ignored   %s [%2d / %2d]'
                                       % (tag, val, rate))
                    return

            if log: log.warning('raise_on triggered %s [%2d / %2d]',
                                  tag, val, rate)
            elif _verb:   print('raise_on triggered %s [%2d / %2d]'
                               % (tag, val, rate))

            # reset counter and raise exception
            raise RuntimeError('raise_on for %s [%s]' % (tag, val))


# ------------------------------------------------------------------------------
#
def attach_pudb(log=None):

    # need to move here to avoid circular import
    from .threads import gettid

    host = '127.0.0.1'
  # host = gethostip()
    tid  = gettid()
    port = tid + 10000

    if log:
        log.info('debugger open: telnet %s %d', host, port)

    print('debugger open: telnet %s %d' % (host, port))

    try:
        import pudb                                      # pylint: disable=E0401
        from   pudb.remote import set_trace              # pylint: disable=E0401

        pudb.DEFAULT_SIGNAL = signal.SIGALRM

        set_trace(host=host, port=port, term_size=(200, 50))

    except Exception as e:
        if log:
            log.exception('failed to attach pudb')
        else:
            print('failed to attach pudb (%s)' % repr(e))


# ------------------------------------------------------------------------------
#
_SNIPPET_PATHS = ['%s/.radical/snippets/' % os.environ.get('RADICAL_BASE', '/tmp')]


def add_snippet_path(path):
    '''
    add a path to the search path for dynamically loaded python snippets
    (see `ru.get_snippet()`).
    '''

    if 'RADICAL_DEBUG' in os.environ:

        if path not in _SNIPPET_PATHS:
            _SNIPPET_PATHS.append(path)


# ------------------------------------------------------------------------------
#
def get_snippet(sid):
    '''
    RU exposes small python snippets for runtime code insertion.  The usage is
    intended as follows:

      * a programmer implements a class
      * for some experiment or test, that class's behavior must be controled at
        runtime.
      * in all places where such an adaptation is expected to take place, the
        programmer inserts a hook like this:

            exec(ru.get_snippet('my_class.init_hook'))

      * this will trigger RU to search for python files of the name
        `my_class.init_hook.py` in `$RADICAL_BASE/.radical/snippets/` (default),
        and return their content for injection.

    The snippet search path can be extended by calling.

        ru.add_snippet_path(path)

    The `RADICAL_DEBUG` environment variable needs to be set for this method to
    do anything.  A snippet can use the following literal strinfgs which will be
    replaced by their actual values:

        '###SNIPPET_FILE###'  - filename from which snippet was loaded
        '###SNIPPET_PATH###'  - path in which the snippet file is located
        '###SNIPPET_ID###'    - the sid string used to identify the snippet
    '''

    if 'RADICAL_DEBUG' in os.environ:

        for path in _SNIPPET_PATHS:

            fname = '%s/%s.py' % (path, sid)

            try:
                with ru_open(fname, 'r') as fin:
                    snippet = fin.read()
                    snippet = snippet.replace('###SNIPPET_FILE###', fname)
                    snippet = snippet.replace('###SNIPPET_PATH###', path)
                    snippet = snippet.replace('###SNIPPET_ID###',   sid)
                    return snippet
            except:
                pass

    return 'None'


# ------------------------------------------------------------------------------
#
class DebugHelper(object):
    '''
    When instantiated, and when 'RADICAL_DEBUG' is set in the environment, this
    class will install a signal handler for SIGUSR1.  When that signal is
    received, a stacktrace for all threads is printed to stdout.
    We also check if SIGINFO is available, which is generally bound to CTRL-T.

    Additionally, a call to 'dh.fs_block(info=None)' will create a file system
    based barrier: it will create a unique file in /tmp/ (based on 'name' if
    given), and dump the stack trace and any 'info' into it.  It then waits
    until that file has changed (touched or removed etc), and then returns.
    The wait is a simple pull based 'os.stat()' (once per sec).
    '''

    # --------------------------------------------------------------------------
    #
    def __init__(self, name=None, info=None):
        '''
        name: string to identify fs barriers
        info: static info to dump into fs barriers
        '''

        self.name   = name
        self.info   = info
        self.locks  = dict()
        self.rlocks = dict()

        if not self.name:
            self.name = str(id(self))

        if 'MainThread' not in mt.current_thread().name:
            # python only supports signals in main threads :-/
            return

        if 'RADICAL_DEBUG' in os.environ:
            signal.signal(signal.SIGUSR1, print_stacktraces)  # signum 30
            signal.signal(signal.SIGQUIT, print_stacktraces)  # signum  3

            try:
                assert signal.SIGINFO
                signal.signal(signal.SIGINFO, print_stacktraces)  # signum 29

            except AttributeError:  # stack unwind in progress
                pass


    # --------------------------------------------------------------------------
    #
    def register_lock(self, name, lock):
        assert name not in self.locks, name
        self.locks[name] = lock


    # --------------------------------------------------------------------------
    #
    def register_rlock(self, name, rlock):
        assert name not in self.rlocks, name
        self.rlocks[name] = rlock


    # --------------------------------------------------------------------------
    #
    def unregister_lock(self, name):
        assert name in self.locks, name
        del self.locks[name]


    # --------------------------------------------------------------------------
    #
    def unregister_rlock(self, name):
        assert name in self.rlocks, name
        del self.rlocks[name]


    # --------------------------------------------------------------------------
    #
    def fs_block(self, info=None):
        '''
        Dump state, info in barrier file, and wait for it tou be touched or
        read or removed, then continue.  Leave no trace.
        '''

        if 'RADICAL_DEBUG' not in os.environ:
            return

        try:
            pid = os.getpid()
            tid = mt.currentThread().ident

            fb  = '/tmp/ru.dh.%s.%s.%s' % (self.name, pid, tid)
            fd  = ru_open(fb, 'w+')

            fd.seek(0,0)
            fd.write('\nSTACK TRACE:\n%s\n%s\n' % (time.time(), get_trace()))
            fd.write('\nSTATIC INFO:\n%s\n\n' % pprint.pformat(self.info))
            fd.write('\nINFO:\n%s\n\n' % pprint.pformat(info))
            fd.flush()

            new = os.stat(fb)
            old = new

            while old == new:
                new = os.stat(fb)
                time.sleep(0.1)

        except:
            # we don't care (much)...
            pass

        finally:
            if fd : fd.close()
            try   : os.unlink(fb)
            except: pass


# ------------------------------------------------------------------------------
#
_debug_helper = None
if 'RADICAL_DEBUG_HELPER' in os.environ:
    if not _debug_helper:
        _debug_helper = DebugHelper()


# ------------------------------------------------------------------------------
#
class Lock(object):

    # --------------------------------------------------------------------------
    #
    def __init__(self, name=None):

        self.lock  = mt.Lock()
        self.owner = None
        self.waits = list()
        self.name  = name

        if not self.name:
            self.name = generate_id('lock')

        if _debug_helper:
            _debug_helper.register_lock(self.name, self)


    def __enter__(self):
        self.acquire()

    def __exit__(self, a, b, c):
        self.release()


    # --------------------------------------------------------------------------
    #
    def acquire(self, blocking=True):

        self.waits.append(get_thread_name())
        ret = self.lock.acquire(blocking=blocking)

        if ret is not False:
            self.owner = get_thread_name()

        self.waits.pop()

        return ret


    # --------------------------------------------------------------------------
    #
    def release(self):

        self.lock.release()
        self.owner = None


# ------------------------------------------------------------------------------
#
class RLock(object):

    # --------------------------------------------------------------------------
    #
    def __init__(self, name=None):

        self.lock  = mt.RLock()
        self.owner = None
        self.waits = list()
        self.name  = name

        if not self.name:
            self.name = generate_id('rlock')

        if _debug_helper:
            _debug_helper.register_rlock(self.name, self)


    def __enter__(self):
        self.acquire()

    def __exit__(self, a, b, c):
        self.release()


    # --------------------------------------------------------------------------
    #
    def acquire(self, blocking=True):

        self.waits.append(get_thread_name())
        ret = self.lock.acquire(blocking=blocking)

        if ret is not False:
            self.owner = get_thread_name()

        self.waits.pop()
        return ret


    # --------------------------------------------------------------------------
    #
    def release(self):

        self.lock.release()
        self.owner = None

# ------------------------------------------------------------------------------




# Copyright 2009 Google Inc.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# Licensed to the PSF under a Contributor Agreement.
#
# Author: Gregory P. Smith <greg@krypto.org>
#
# pylint: disable=protected-access

'''
This module provides code to setup appropriate atfork() calls within the
Python standard library in order to make it safe for use in programs that
mix fork() and threads.

Import this and call the appropriate fixers very early in your program's
initialization sequence, preferrably -before- importing the modules you
need to fix.

Provided fixers:

    fix_logging_module()

TODO(gps): Audit more of the stdlib and provide necessary fixers.
In 2.4.5 the following additional stdlib modules use locks:

  threading.Condition and similar classes could use it.
  Queue
  cookielib
  mimetools
  _strptime
'''

import os
import sys

from .atfork import atfork


# ------------------------------------------------------------------------------
#
def _warn(msg):

    import warnings

    def custom_formatwarning(msg, *args, **kwargs):
        return 'WARNING: %s\n' % str(msg)

    orig_formatwarning     = warnings.formatwarning
    warnings.formatwarning = custom_formatwarning
    warnings.warn(msg)
    warnings.formatwarning = orig_formatwarning


# ------------------------------------------------------------------------------
#
class Error(Exception):
    pass


# ------------------------------------------------------------------------------
#
def fix_logging_module():

    # monkeypatching can be disabled by setting RADICAL_UTILS_NO_ATFORK
    if 'RADICAL_UTILS_NO_ATFORK' in os.environ:
        return

    logging = sys.modules.get('logging')
    # Prevent fixing multiple times as that would cause a deadlock.
    if logging and getattr(logging, 'fixed_for_atfork', None):
        return

    if logging:
      # if os.environ.get('RADICAL_DEBUG'):
      #     warnings.warn('logging module already imported before fixup.')
        pass

    import logging
    if logging.getLogger().handlers:
        # We could register each lock with atfork for these handlers but if
        # these exist, other loggers or not yet added handlers could as well.
        # Its safer to insist that this fix is applied before logging has been
        # configured.
        _warn('Import `radical` modules before `logging` to avoid the '
              'application to deadlock on `fork()`!')

    logging._acquireLock()
    try:
        def fork_safe_createLock(self):
            self._orig_createLock()
            atfork(self.lock.acquire,
                   self.lock.release, self.lock.release)

        # Fix the logging.Handler lock (a major source of deadlocks).
        logging.Handler._orig_createLock = logging.Handler.createLock
        logging.Handler.createLock = fork_safe_createLock

        # Fix the module level lock.
        atfork(logging._acquireLock,
               logging._releaseLock, logging._releaseLock)

        logging.fixed_for_atfork = True
    finally:
        logging._releaseLock()

# ------------------------------------------------------------------------------




#!/usr/bin/env python

__author__    = "Radical.Utils Development Team (Andre Merzky, Ole Weidner)"
__copyright__ = "Copyright 2013, RADICAL@Rutgers"
__license__   = "MIT"


'''
And example on using the radical.utils.config tools.

This example will read config options fomr $HOME/.examples.cfg::

    [config]
    casing   = upper
    excluded = sparks,pftools

    [sp3.cd]
    exe  = /usr/local/bin/sp3
    args = no,idea

That setting can be overwritting via $EXAMPLE_CONFIG_CASING.
'''

import radical.utils.config  as ruc


# ------------------------------------------------------------------------------
#
# a set of pre-defined options
#
_config_options = {'casing'  : '${EXAMPLE_CONFIG_CASING:default}',
                   'excluded': []}

_sp3_options    = {'exe'     : '/usr/bin/sp3',
                   'args'    : []}


# ------------------------------------------------------------------------------
#
class FancyEcho(ruc.Configurable):
    """
    This example will evaluate the given configuration, and
    """

    # --------------------------------------------------------------------------
    #
    def __init__(self):

        ruc.Configurable.__init__(self, 'radical.utils')
        self._cfg.merge(_config_options)
        self._cfg.merge(_sp3_options)

        self._mode = self._cfg['casing']
        print("mode: %s" % self._mode)

        self._excl = self._cfg['excluded']
        print("excl: %s" % type(self._excl))
        print("excl: %s" % self._excl)


        if 'sp3' not in self._excl:
            print('running sp3')

        if  'sparks' not in self._excl:
            print('running sparks')

        if  'pftools' not in self._excl:
            print('running pftools')

        # use sp3 configuration
        print(self._cfg['exe'])
        print(self._cfg['args'])


    # --------------------------------------------------------------------------
    #
    def echo(self, source):

        target = ""

        if  self._mode == 'default':
            target = source

        elif self._mode == 'lower':
            target = source.lower()

        elif self._mode == 'upper':
            target = source.upper()

        else:
            target = 'cannot handle mode %s' % self._mode

        return target


# ------------------------------------------------------------------------------
#

fc  = FancyEcho()
src = ''

while 'quit' != src:
    src = input('> ')
    tgt = fc.echo(src)
    print('  ' + tgt)


# ------------------------------------------------------------------------------
#






#!/usr/bin/env python3

import time

import threading     as mt

import radical.utils as ru

CHANNEL = 'channel'
TOPIC   = 'test'


# ------------------------------------------------------------------------------
#
def pub(uid, url, topic, n):

    pub = ru.zmq.Publisher(topic, url)
    for i in range(n):
        pub.put(topic, ru.as_bytes('%s: message %d' % (uid, i)))
        print('%s: message %d' % (uid, i))
        time.sleep(0.1)

    pub.put(topic, ru.as_bytes('%s: STOP' % uid))

    print('%s: done' % uid)


# ------------------------------------------------------------------------------
#
def sub(uid, url, topic):

    cont = True
    sub  = ru.zmq.Subscriber(topic, url)
    sub.subscribe(topic)

    while cont:
        topic, msg = sub.get()
        print('%s: %s' % (uid, ru.as_string(msg)))

        if 'STOP' in ru.as_string(msg):
            cont = False

    print('%s: done' % uid)


# ------------------------------------------------------------------------------
#
def main():

    bridge = ru.zmq.PubSub(CHANNEL)
    bridge.start()
    time.sleep(1)

    threads = list()


    # start some subscribers and publishers
    for i in range(3):
        t = mt.Thread(target=sub, args=('sub.%d' % i, bridge.addr_sub, TOPIC))
        t.daemon = True
        t.start()
        threads.append(t)

    for i in range(4):
        t = mt.Thread(target=pub, args=('pub.%d' % i, bridge.addr_pub, TOPIC, 5))
        t.daemon = True
        t.start()
        threads.append(t)


    # wait for completion
    for t in threads:
        t.join()


# ------------------------------------------------------------------------------
#
if __name__ == '__main__':

    main()


# ------------------------------------------------------------------------------





import os
import sys
import string

# import colorama as c

from .misc    import get_env_ns as ru_get_env_ns
from .misc    import ru_open
from .config  import DefaultConfig


# ------------------------------------------------------------------------------
#
def _open(target):

    try:
        os.makedirs(os.path.abspath(os.path.dirname(target)))
    except OSError:
        pass  # exists

    return ru_open(target, 'w')


# ------------------------------------------------------------------------------
#
class Reporter(object):

    # COLORS = {'white'       : c.Style.BRIGHT    + c.Fore.WHITE   ,
    #           'yellow'      : c.Style.BRIGHT    + c.Fore.YELLOW  ,
    #           'green'       : c.Style.BRIGHT    + c.Fore.GREEN   ,
    #           'blue'        : c.Style.BRIGHT    + c.Fore.BLUE    ,
    #           'cyan'        : c.Style.BRIGHT    + c.Fore.CYAN    ,
    #           'red'         : c.Style.BRIGHT    + c.Fore.RED     ,
    #           'magenta'     : c.Style.BRIGHT    + c.Fore.MAGENTA ,
    #           'black'       : c.Style.BRIGHT    + c.Fore.BLACK   ,
    #           'darkwhite'   : c.Style.DIM       + c.Fore.WHITE   ,
    #           'darkyellow'  : c.Style.DIM       + c.Fore.YELLOW  ,
    #           'darkgreen'   : c.Style.DIM       + c.Fore.GREEN   ,
    #           'darkblue'    : c.Style.DIM       + c.Fore.BLUE    ,
    #           'darkcyan'    : c.Style.DIM       + c.Fore.CYAN    ,
    #           'darkred'     : c.Style.DIM       + c.Fore.RED     ,
    #           'darkmagenta' : c.Style.DIM       + c.Fore.MAGENTA ,
    #           'darkblack'   : c.Style.DIM       + c.Fore.BLACK   ,
    #           'reset'       : c.Style.RESET_ALL + c.Fore.RESET
    #       }

    COLORS = {'reset'        : '\033[39m',
              'black'        : '\033[30m',
              'red'          : '\033[31m',
              'green'        : '\033[32m',
              'yellow'       : '\033[33m',
              'blue'         : '\033[34m',
              'magenta'      : '\033[35m',
              'cyan'         : '\033[36m',
              'lightgray'    : '\033[37m',
              'darkgray'     : '\033[90m',
              'lightred'     : '\033[91m',
              'lightgreen'   : '\033[92m',
              'lightyellow'  : '\033[93m',
              'lightblue'    : '\033[94m',
              'lightmagenta' : '\033[95m',
              'lightcyan'    : '\033[96m',
              'white'        : '\033[97m'}

    MODS   = {'reset'        : '\033[0m',
              'bold'         : '\033[1m',
              'underline'    : '\033[4m',
              'blink'        : '\033[5m',
              'inverse'      : '\033[7m',
              ''             : ''}


    # define terminal colors and other output options
    TITLE    = 'bold lightblue'
    HEADER   = 'bold lightyellow'
    INFO     = 'lightblue'
    IDLE     = 'lightwhite'
    PROGRESS = 'lightwhite'
    OK       = 'lightgreen'
    WARN     = 'lightyellow'
    ERROR    = 'lightred'

    EMPTY    = ''
    DOTTED   = '.'
    SINGLE   = '-'
    DOUBLE   = '='
    HASHED   = '#'


    # --------------------------------------------------------------------------
    #
    def __init__(self, name, ns=None, path=None, targets=None, enabled=None):
        '''
        settings.style:
          E : empty line
          T : tabulator
          L : line of line segments
          M : text to report
        '''

        ru_def = DefaultConfig()

        self._name = name
        if not ns:
            ns = name

        # check if this profile is enabled via an env variable
        if enabled is not None:
            self._enabled = bool(enabled)
        else:
            self._enabled = str(ru_def['report']).lower()

            if ru_get_env_ns('report', ns) is not None:
                self._enabled = str(ru_get_env_ns('report', ns)).lower()

            if self._enabled in ['0', 'false', 'off', 'none']:
                self._enabled = False
            else:
                self._enabled = True


        if not self._enabled:
            # disabled
            return


        self._use_color = ru_get_env_ns('report_color', ns, default='True')
        if self._use_color.lower() in ['0', 'false', 'off']:
            self._use_color = False
        else:
            self._use_color = True


        self._use_anime = ru_get_env_ns('report_anime', ns, default='True')
        if self._use_anime.lower() in ['0', 'false', 'off']:
            self._use_anime = False
        else:
            self._use_anime = True


        self._line_len = int(ru_get_env_ns('report_llen', ns, default=80))


        if not path:
            path = os.getcwd()

        if not targets:
            targets = ru_get_env_ns('report_tgt', ns)
            if not targets:
                targets = ru_def['report_tgt']

        if isinstance(targets, str):
            targets = targets.split(',')

        if not isinstance(targets, list):
            targets = [targets]

        if '/' in name:
            try:
                os.makedirs(os.path.normpath(os.path.dirname(name)))
            except OSError:
                pass  # dir exists

        self._prog_tgt = None
        self._prog_cnt = 0
        self._prog_pos = 0
        self._prog_off = 0

        C = 'color'
        S = 'style'
        G = 'segment'

        self._pos      = 0
        self._settings = {
            'title'    : {C: self.TITLE    , S: 'ELMLE' , G: self.DOUBLE},
            'header'   : {C: self.HEADER   , S: 'ELME'  , G: self.SINGLE},
            'info'     : {C: self.INFO     , S: 'M'     , G: self.EMPTY },
            'idle'     : {C: self.IDLE     , S: 'M'     , G: self.EMPTY },
            'progress' : {C: self.PROGRESS , S: 'M'     , G: self.EMPTY },
            'ok'       : {C: self.OK       , S: 'M'     , G: self.EMPTY },
            'warn'     : {C: self.WARN     , S: 'M'     , G: self.EMPTY },
            'error'    : {C: self.ERROR    , S: 'M'     , G: self.EMPTY },
            'plain'    : {C: ''            , S: 'M'     , G: self.EMPTY },
        }

        if not self._use_color:
            for k in self._settings:
                self._settings[k]['color'] = ''

        self._idle_sequence = '/-\\|'
        self._idle_pos      = dict()
        self._idle_count    = 0
        self._streams       = list()

        for t in targets:
            if   t in ['0', 'null']       : continue
            elif t in ['-', '1', 'stdout']: h = sys.stdout
            elif t in ['=', '2', 'stderr']: h = sys.stderr
            elif t in ['.']               : h = _open("%s/%s.rep" % (path,name))
            elif t.startswith('/')        : h = _open(t)
            else                          : h = _open("%s/%s"     % (path,t))

            self._streams.append(h)


    # --------------------------------------------------------------------------
    #
    def _out(self, color, msg, count=None):

        if not self._enabled:
            return

        if self._use_color:
            color_mod = ''
            if  ' ' in color:
                color_mod, color = color.split(' ', 2)

            color     = self.COLORS.get(color.lower(), '')
            color_mod = self.MODS.get(color_mod.lower(), '')

            color += color_mod


        # make sure we count tab length on line start correctly
        msg = msg.replace('\n\t', '\n        ')

        # make sure we don't extent a long line further
        if self._pos >= (self._line_len) and msg and msg[0] != '\n':
            while msg[0] == '\b':
                msg = msg[1:]
            if count:
                msg = ' %6d\n        %s' % (count, msg)
            else:
                msg = '\n        %s' % msg

        # special control characters:
        #
        #   * '>>' will, at it's place, insert sufficient spaces to make the
        #     remainder of the string right-aligned.  Only one >> is
        #     interpreted, linebreaks before it are ignored
        #
        #   * '<<' will insert a line break if the position is not already on
        #     the beginning of a line
        #
        slash_f  = msg.find('>>')
        if slash_f >= 0:
            copy   = msg[slash_f + 1:].strip()
            spaces = self._line_len - self._pos - len(copy) + 1  # '>>'
            spaces = max(0, spaces)
            msg    = msg.replace('>>', spaces * ' ')

        slash_cr = msg.find('<<')
        if slash_cr >= 0:
            if self._pos + slash_cr > 0:
                spaces = self._line_len - self._pos - 1
                msg    = msg.replace('<<', '%s\\\n' % (spaces * ' '))
            else:
                msg = msg.replace('<<', '')

        mlen  = len([x for x in msg if x in string.printable])
        mlen -= msg.count('\b')

        # find the last \n and then count how many chars we are writing after it
        slash_n = msg.rfind('\n')

        if slash_n >= 0:
            self._pos = mlen - slash_n - 1
        else:
            self._pos += mlen

        for stream in self._streams:
            if self._use_color:
                stream.write(color)
            stream.write(msg)
            if self._use_color:
                stream.write(self.COLORS['reset'])
                stream.write(self.MODS['reset'])
            try:
                stream.flush()
            except Exception:
                pass


    # --------------------------------------------------------------------------
    #
    def _format(self, msg, settings=None):

        if not self._enabled:
            return

        if not msg:
            msg = ''

        if not settings:
            settings = dict()

        color   = settings.get('color',   '')
        style   = settings.get('style',   'M')
        segment = settings.get('segment', '')

        for c in style:

            if  c == 'M':
                self._out(color, "%s" % msg)

            if  c == 'T':
                self._out(color, "\t")

            elif c == 'E':
                self._out(color, "\n")

            elif c == 'L':
                if segment:
                    self._out(color, "%s\n" % (self._line_len * segment))


    # --------------------------------------------------------------------------
    #
    def set_style(self, which, color=None, style=None, segment=None):

        if not self._enabled:
            return

        if which not in self._settings:
            raise LookupError('reporter does not support style "%s"' % which)

        settings = self._settings[which]

        if color  : settings['color']   = color
        if style  : settings['style']   = style
        if segment: settings['segment'] = segment


    # --------------------------------------------------------------------------
    #
    def title(self, title):

        if not self._enabled:
            return

        if not title:
            title = self._name

        if not title:
            return

        fmt = " %%-%ds\n" % (self._line_len - 1)
        self._format(fmt % title, self._settings['title'])


    # --------------------------------------------------------------------------
    #
    def header(self, msg=''):

        if not self._enabled:
            return

        if msg:
            fmt = "%%-%ds\n" % self._line_len
            msg = fmt % msg

        self._format(msg, self._settings['header'])


    # --------------------------------------------------------------------------
    #
    def info(self, msg=''):

        if not self._enabled:
            return

        self._format(msg, self._settings['info'])


    # --------------------------------------------------------------------------
    #
    def idle(self, c=None, mode=None, color=None, idle_id=None):

        if not self._enabled:
            return

        if not self._use_anime:
            return

        if not idle_id:
            idle_id = 'default'

        if color: col = self._settings[color]['color']
        else    : col = self._settings['idle']['color']

        idx = 0
        if mode == 'start':
            self._out(col, 'O')
            self._idle_count = 0

        elif mode == 'stop':
            self._out(col, '\b %6d' % self._idle_count)

        else:
            if not c:
                idx  = self._idle_pos.get(idle_id, 0)
                c    = self._idle_sequence[idx % len(self._idle_sequence)]
                idx += 1
                self._out(col, '\b%s' % c)
            else:
                idx += 1
                self._idle_count += 1
                self._out(col, '\b%s|' % c, count=self._idle_count)

        self._idle_pos[idle_id] = idx


    # --------------------------------------------------------------------------
    #
    def progress_tgt(self, tgt=None, label=None):

        if not self._enabled:
            return

        if label:
            front = '%-6s: ' % label
        else:
            front = '%-6d: ' % tgt

        self._prog_len = self._line_len - len(front)
        self._prog_tgt = tgt
        self._prog_cnt = 0
        self._prog_pos = 0

        self._format(front, self._settings['progress'])


    # --------------------------------------------------------------------------
    #
    def progress_done(self):

        if not self._enabled:
            return

        self._prog_tgt = None
        self._prog_cnt = 0
        self._prog_pos = 0
        self._format('\n', self._settings['progress'])


    # --------------------------------------------------------------------------
    #
    def progress(self, msg=None):

        if not self._enabled:
            return

        if self._prog_tgt:
            self._prog_cnt += 1

            val  = int(self._prog_cnt * self._prog_len / self._prog_tgt)

            while val > self._prog_pos:
                self._prog_pos += 1
                self._format('#', self._settings['progress'])

        else:

            if not msg:
                msg = '.'

            self._format(msg, self._settings['progress'])


    # --------------------------------------------------------------------------
    #
    def ok(self, msg=''):

        if not self._enabled:
            return

        self._format(msg, self._settings['ok'])


    # --------------------------------------------------------------------------
    #
    def warn(self, msg=''):

        if not self._enabled:
            return

        self._format(msg, self._settings['warn'])


    # --------------------------------------------------------------------------
    #
    def error(self, msg=''):

        if not self._enabled:
            return

        self._format(msg, self._settings['error'])


    # --------------------------------------------------------------------------
    #
    def exit(self, msg='', exit_code=0):

        if not self._enabled:
            return

        self.error(msg)
        sys.exit(exit_code)


    # --------------------------------------------------------------------------
    #
    def plain(self, msg=''):

        if not self._enabled:
            return

        self._format(msg)


# ------------------------------------------------------------------------------





__copyright__ = "Copyright 2013-2014, http://radical.rutgers.edu"
__license__   = "MIT"


# ------------------------------------------------------------------------------
#
"""
Using the RU logging module can lead to deadlocks when used in multiprocess and
multithreaded environments.  This is due to the native python logging library
not being thread save.  See [1] for a Python bug report on the topic from 2009.
As of 2016, a patch is apparently submitted, but not yet accepted, for Python
3.5.  So we have to solve the problem on our own.

The fix is encoded in the 'after_fork()' method.  That method *must* be called
immediately after a process fork.  A fork will carry over all locks from the
parent process -- but it will not carry over any threads.  That means that, if
a fork happens while a thread owns the lock, the child process will start with
a locked lock -- but since there are no threads, there is actually nobody who
can unlock the lock, thus badaboom.

Since after the fork we *know* the logging locks should be unset (after all, we
are not in a logging call right now, are we?), we pre-emtively unlock them here.
But, to do so we need to know what locks exist in the first place.  For that
purpose, we create a process-singleton of all the loggers we hand out via
'get_logger()'.

Logging locks are 'threading.RLock' instances.  As such they can be locked
multiple times (from within the same thread), and we have to unlock them that
many times.  We use a shortcut, and create a new, unlocked lock.
"""


# ------------------------------------------------------------------------------
#
# NOTE: ForkingPickler does not like lambdas nor local functions, thus use
#       a module level function to disable loggers.
#
def _noop(*args, **kwargs):
    pass


# ------------------------------------------------------------------------------
#
import os
import sys
import threading
import colorama
import logging

from typing import Dict

from   .atfork    import atfork
from   .misc      import get_env_ns       as ru_get_env_ns
from   .modules   import import_module    as ru_import_module
from   .config    import DefaultConfig
from   .singleton import Singleton


CRITICAL = logging.CRITICAL
ERROR    = logging.ERROR
WARNING  = logging.WARNING
WARN     = logging.WARNING
INFO     = logging.INFO
DEBUG    = logging.DEBUG
OFF      = 60


# ------------------------------------------------------------------------------
#
class _LoggerRegistry(object, metaclass=Singleton):

    from .singleton import Singleton

    def __init__(self):
        self._registry = list()

    def add(self, logger):
        self._registry.append(logger)

    def release_all(self):
        for logger in self._registry:
            while logger:
                for handler in logger.handlers:
                    handler.lock = threading.RLock()
                  # handler.reset()
                logger = logger.parent

    def close_all(self):

        for logger in self._registry:
            while logger:
                for handler in logger.handlers:
                    handler.close()
                    logger.removeHandler(handler)
                logger = logger.parent
        self._registry = list()


# ------------------------------------------------------------------------------
#
_logger_registry = _LoggerRegistry()


# ------------------------------------------------------------------------------
def _after_fork():

    _logger_registry.release_all()
    logging._lock = threading.RLock()         # pylint: disable=protected-access


# ------------------------------------------------------------------------------
#
def _atfork_prepare():
    pass


# ------------------------------------------------------------------------------
#
def _atfork_parent():
    pass


# ------------------------------------------------------------------------------
#
def _atfork_child():
    _after_fork()


# ------------------------------------------------------------------------------
#
atfork(_atfork_prepare, _atfork_parent, _atfork_child)


# ------------------------------------------------------------------------------
#
class ColorStreamHandler(logging.StreamHandler):
    """
    A colorized output SteamHandler
    """

    # --------------------------------------------------------------------------
    #
    def __init__(self, target):

        logging.StreamHandler.__init__(self, target)
        self._tty    = self.stream.isatty()
        self._term   = getattr(self, 'terminator', '\n')
        self.colours = {'DEBUG'    : colorama.Fore.CYAN,
                        'INFO'     : colorama.Fore.GREEN,
                        'WARN'     : colorama.Fore.YELLOW,
                        'WARNING'  : colorama.Fore.YELLOW,
                        'ERROR'    : colorama.Fore.RED,
                        'CRITICAL' : colorama.Back.RED + colorama.Fore.WHITE,
                        'RESET'    : colorama.Style.RESET_ALL + self._term
                       }


    # --------------------------------------------------------------------------
    #
    def emit(self, record):

        # only write in color when using a tty
        if self._tty:
            self.stream.write('%s%s%s' % (self.colours[record.levelname],
                                          self.format(record),
                                          self.colours['RESET']))
        else:
            self.stream.write(self.format(record) + self._term)
        self.stream.flush()


# ------------------------------------------------------------------------------
#
class FSHandler(logging.FileHandler):

    def __init__(self, target):

        try:
            os.makedirs(os.path.abspath(os.path.dirname(target)))
        except:
            pass  # exists

        logging.FileHandler.__init__(self, target, delay=True)


# ------------------------------------------------------------------------------
#
class Logger(object):

    '''
    Logger documentation
    '''

    # --------------------------------------------------------------------------
    #
    def __init__(self, name, ns=None, path=None, targets=None, level=None,
                 debug=None, verbose=False):
        """
        Get a logging handle.

        `name`    is used to identify log entries on this handle.
        `ns`      is used as environment name space to check for the log level
        `targets` is a comma separated list (or Python list) of specifiers,
                  where specifiers are:

                    `0`      : /dev/null
                    `null`   : /dev/null
                    `-`      : stdout
                    `1`      : stdout
                    `stdout` : stdout
                    `=`      : stderr
                    `2`      : stderr
                    `stderr` : stderr
                    `.`      : logfile named ./<name>.log
                    <string> : logfile named <string>

        `path`    file system location to write logfiles to (created as needed)
        `level`   log level (DEBUG, INFO, WARNING, ERROR, CRITICAL, OFF)
        `debug`   debug level (0-9)

        If `ns` is, for example, set to `radical.utils`, then the following
        environment variables are evaluated:

            RADICAL_UTILS_LOG_LVL
            RADICAL_LOG_LVL

            RADICAL_UTILS_LOG_TGT
            RADICAL_LOG_TGT

        The first found variable of each pair is then used for the respective
        settings.
        """

        if name is None:
            raise ValueError('logger name must be specified and not `None`')

        self._name        = name
        self._ns          = ns
        self._path        = path
        self._targets     = targets
        self._level       = level
        self._debug_level = debug
        self._debug       = debug
        self._verbose     = verbose
        self._num_level   = 0
        self._logger      = None


    # --------------------------------------------------------------------------
    #
    def _ensure_handler(self):

        if self._logger:
            return

        ru_def = DefaultConfig()

        # make sure handlers are attached.  If they are not, attach them!
        self._logger = logging.getLogger(self._name)
        self._logger.propagate = False   # let messages not trickle upward
        self._logger.name      = self._name

        # otherwise configure this logger
        if not self._path:
            self._path = ru_def['log_dir']

        if not self._ns:
            self._ns = self._name

        if not self._targets:
            self._targets = ru_get_env_ns('log_tgt', self._ns)
            if not self._targets:
                self._targets = ru_def['log_tgt']

        if isinstance(self._targets, str):
            self._targets = self._targets.split(',')

        if not isinstance(self._targets, list):
            self._targets = [self._targets]

        if not self._level:
            self._level = ru_get_env_ns('log_lvl', self._ns)

        if not self._level:
            # backward compatibility
            self._level = ru_get_env_ns('verbose', self._ns)

        if not self._level:
            self._level = ru_def['log_lvl']

        if self._level in [OFF, 'OFF']:
            self._targets = ['null']

        try:
            self._level = int(self._level)
        except:
            pass

        self._debug_level = self._debug or 0
        if isinstance(self._level, int):
            if self._level < 10:
                self._debug_level = 10 - self._level
                self._level = 'DEBUG'

        elif self._level.upper().startswith('DEBUG_'):
            self._debug_level = int(self._level.split('_', 1)[1])
            self._level = 'DEBUG'

        if self._level != 'DEBUG':
            self._debug_level = 0


        # translate numeric levels into upper case symbolic ones
        levels   = {'60' : 'OFF',
                    '50' : 'CRITICAL',
                    '40' : 'ERROR',
                    '30' : 'WARNING',
                    '20' : 'INFO',
                    '10' : 'DEBUG',
                     '0' :  ru_def['log_lvl']}


        self._level   = levels.get(str(self._level), str(self._level)).upper()
        self._warning = None
        if self._level not in list(levels.values()):
            self._warning = "invalid loglevel '%s', use '%s'" \
                                      % (self._level, ru_def['log_lvl'])
            self._level   = ru_def['log_lvl']

        formatter = logging.Formatter('%(created).3f : '
                                      '%(name)-20s : '
                                      '%(process)-5d : '
                                      '%(thread)-5d : '
                                      '%(levelname)-8s : '
                                      '%(message)s')

      # print('%-30s -> %-10s %d' % (name, level, debug_level))

        # add a handler for each targets (using the same format)
        if not self._logger.handlers:
            p = self._path
            n = self._name
            for t in self._targets:

                if   t in ['0', 'null']       : h = logging.NullHandler()
                elif t in ['-', '1', 'stdout']: h = ColorStreamHandler(sys.stdout)
                elif t in ['=', '2', 'stderr']: h = ColorStreamHandler(sys.stderr)
                elif t in ['.']               : h = FSHandler("%s/%s.log" % (p, n))
                elif t.startswith('/')        : h = FSHandler(t)
                else                          : h = FSHandler("%s/%s"     % (p, t))

                h.setFormatter(formatter)
                h.name = self._logger.name
                self._logger.addHandler(h)

            if self._level != 'OFF':
                self._logger.setLevel(self._level)

            if self._warning:
                self._logger.warning(self._warning)

        # if `name` points to module, try to log its version info
        if self._verbose:

            self._ensure_handler()

            try:
                self._logger.info("%-20s version: %s", 'python.interpreter',
                                  ' '.join(sys.version.split()))

                mod = ru_import_module(self._name)
                if hasattr(mod, 'version_detail'):
                    self._logger.info("%-20s version: %s",
                                      self._name, getattr(mod, 'version_detail'))

                elif hasattr(mod, 'version'):
                    self._logger.info("%-20s version: %s",
                                      self._name, getattr(mod, 'version'))
            except:
                pass

            # also log pid and tid
            try:
                self._logger.info("%-20s pid/tid: %s/%s", '', os.getpid(),
                            threading.current_thread().name)
            except:
                pass


        # keep the handle a round, for cleaning up on fork
        _logger_registry.add(self._logger)

        self._numerics : Dict[str, int] = {'off'     : 60,
                                           'critical': 50,
                                           'error'   : 40,
                                           'warning' : 30,
                                           'info'    : 20,
                                           'debug'   : 10}

        # store properties
        self._num_level = self._numerics.get(self._level.lower(), 0)

        if self._level == 'DEBUG':
            self._num_level -= self._debug_level

        # backward compatibility
        self._logger.warn = self._logger.warning

        # treat `self.debug_1()`, `self.debug_2()` etc. the same as
        # `self.debug()` if the respective `debug_level` is set, and ignore
        # otherwise.  All other unknown method calls are forwarded to the nativ
        # logger instance.  This is basically inheritance, but since the logger
        # class has no constructor, we do it this way.

        # disable all loggers
        self.critical = _noop
        self.error    = _noop
        self.warn     = _noop
        self.warning  = _noop
        self.info     = _noop
        self.debug    = _noop
        self.debug_1  = _noop
        self.debug_2  = _noop
        self.debug_3  = _noop
        self.debug_4  = _noop
        self.debug_5  = _noop
        self.debug_6  = _noop
        self.debug_7  = _noop
        self.debug_8  = _noop
        self.debug_9  = _noop

        # enable the ones we are configured for:
        if self._num_level <= 50: self.critical = self._logger.critical
        if self._num_level <= 40: self.error    = self._logger.error
        if self._num_level <= 30: self.warn     = self._logger.warn
        if self._num_level <= 30: self.warning  = self._logger.warning
        if self._num_level <= 20: self.info     = self._logger.info
        if self._num_level <= 10: self.debug    = self._logger.debug
        if self._num_level <=  9: self.debug_1  = self._logger.debug
        if self._num_level <=  8: self.debug_2  = self._logger.debug
        if self._num_level <=  7: self.debug_3  = self._logger.debug
        if self._num_level <=  6: self.debug_4  = self._logger.debug
        if self._num_level <=  5: self.debug_5  = self._logger.debug
        if self._num_level <=  4: self.debug_6  = self._logger.debug
        if self._num_level <=  3: self.debug_7  = self._logger.debug
        if self._num_level <=  2: self.debug_8  = self._logger.debug
        if self._num_level <=  1: self.debug_9  = self._logger.debug



    # --------------------------------------------------------------------------
    #
    @property
    def name(self):
        return self._name

    @property
    def ns(self):
        return self._ns

    @property
    def path(self):
        return self._path

    @property
    def level(self):
        return self._level

    @property
    def debug_level(self):
        return self._debug_level

    @property
    def num_level(self):
        return self._num_level

    @property
    def targets(self):
        return self._targets


    # --------------------------------------------------------------------------
    #
    # All unknown method calls are forwarded to the nativ logger instance.
    # This is basically inheritance, but since the logger class has no
    # constructor, we do it this way.
    #
    def __getattr__(self, name):

        self._ensure_handler()
        try:
            return getattr(self, name)
        except:
            return getattr(self._logger, name)


    # --------------------------------------------------------------------------
    #
    # Add a close method to make sure we can close file handles etc.  This also
    # closes handles on all parent loggers (if those exist).
    #
    def close(self):

        logger = self._logger
        while logger:
            for handler in logger.handlers:
                handler.close()
                logger.removeHandler(handler)
            logger = logger.parent


# ------------------------------------------------------------------------------





__author__    = "Radical.Utils Development Team"
__copyright__ = "Copyright 2016, RADICAL@Rutgers"
__license__   = "MIT"


import os
import sys
import signal
import ctypes
import _thread

import threading    as mt


# ------------------------------------------------------------------------------
#
def get_thread_name():

    if not mt.current_thread():
        return None

    return mt.current_thread().name


# ------------------------------------------------------------------------------
#
def get_thread_id():

    return mt.current_thread().ident


# ------------------------------------------------------------------------------
#
def gettid():
    """
    Python is not able to give us the native thread ID.  We thus use a syscall
    to do so.  Since this is not portable, we fall back to None in case of error
    (Hi MacOS).
    """
    try:
        SYS_gettid = 186
        libc = ctypes.cdll.LoadLibrary('libc.so.6')
        return int(libc.syscall(SYS_gettid))
    except:
        return None


# ------------------------------------------------------------------------------
#
def is_main_thread(t=None):

    if t:
        assert isinstance(t, mt.Thread)
    else:
        t = this_thread()

    return isinstance(t, mt._MainThread)                 # pylint: disable=W0212


# ------------------------------------------------------------------------------
#
def is_this_thread(t):
    '''
    check if the given thread (type: threading.Thread) is the same as the
    current thread of execution.
    '''

    assert isinstance(t, mt.Thread)

    return bool(t == this_thread())


# ------------------------------------------------------------------------------
#
def main_thread():
    '''
    return a handle to the main thread of execution in this process
    '''

    for t in mt.enumerate():
        if isinstance(t, mt._MainThread):                # pylint: disable=W0212
            return t

    assert False, 'main thread not found'


# ------------------------------------------------------------------------------
#
def this_thread():
    '''
    return a handle to the current thread of execution
    '''

    return mt.current_thread()


# ------------------------------------------------------------------------------
#
_signal_lock = mt.Lock()
_signal_sent = dict()


def cancel_main_thread(signame=None, once=False):
    """
    This method will call thread.interrupt_main from any calling subthread.
    That will cause a 'KeyboardInterrupt' exception in the main thread.  This
    can be excepted via `except KeyboardInterrupt`

    The main thread MUST NOT have a SIGINT signal handler installed (other than
    the default handler or SIGIGN), otherwise this call will cause an exception
    in the core python signal handling (see http://bugs.python.org/issue23395).

    The thread will exit after this, via sys.exit(0), and can then be joined
    from the main thread.

    When being called *from* the main thread, no interrupt will be generated,
    but sys.exit() will still be called.  This can be excepted in the code via
    `except SystemExit`.

    Another way to avoid the SIGINT problem is to send a different signal to the
    main thread.  We do so if `signal` is specified.

    After sending the signal, any sub-thread will call sys.exit(), and thus
    finish.  We leave it to the main thread though to decide if it will exit at
    this point or not.  Either way, it will have to handle the signal first.

    If `once` is set to `True`, we will send the given signal at most once.
    This will mitigate races between multiple error causes, specifically during
    finalization.
    """

    if signame: signum = get_signal_by_name(signame)
    else      : signum = None

    with _signal_lock:

        if once:
            if signum in _signal_sent:
                # don't signal again
                return

        if signum:
            # send the given signal to the process to which this thread belongs
            os.kill(os.getpid(), signum)
        else:
            # this sends a SIGINT, resulting in a KeyboardInterrupt.
            # NOTE: see http://bugs.python.org/issue23395 for problems on using
            #       SIGINT in combination with signal handlers!
            try:
                _thread.interrupt_main()
            except TypeError:
                # this is known to be a side effect of `thread.interrup_main()`
                pass

        # record the signal sending
        _signal_sent[signum] = True

    # the sub thread will at this point also exit.
    if not is_main_thread():
        sys.exit()


# ------------------------------------------------------------------------------
# this is the counterpart for the `cancel_main_thread` method above: any main
# thread should call `set_cancellation_handler`, which will set a signal handler
# for `SIGUSR2`, and upon catching it will raise a `KeyboardInterrupt`
# exception, which can be caught by any interested library or application.
#
# RU claims ownership of `SIGUSR2` -- it will complain if any other signal
# handler is installed for that signal.
#
# This method can safely be called multiple times.  This method can be called
# from threads, but it will have no effect then (Python allows signal handlers
# to only be installed in the main thread)
#
# FIXME: `cancel_main_thread()` supports arbitrary signals --
#        `set_cancellation_handler()` should, too.
#
def _sigusr2_handler(signum, frame):
    print('caught sigusr2 (%s)' % os.getpid())
    # we only want to get this exception once, so we unset the signal handler
    # before we raise it
    signal.signal(signal.SIGUSR2, signal.SIG_IGN)
    raise KeyboardInterrupt('sigusr2')


def set_cancellation_handler():

    # this method compares function pointers
    # pylint: disable=W0143

    # check if any handler exists
    old_handler = signal.getsignal(signal.SIGUSR2)
    if old_handler not in [signal.SIG_DFL, signal.SIG_IGN, None] and \
       old_handler != _sigusr2_handler:
        raise RuntimeError('handler for SIGUSR2 is already present')

    try:
        signal.signal(signal.SIGUSR2, _sigusr2_handler)
    except ValueError:
        # this fails in subthreads - ignore
        pass


def unset_cancellation_handler():

    try:
        signal.signal(signal.SIGUSR2, signal.SIG_IGN)
    except ValueError:
        # this fails in subthreads - ignore
        pass


# ------------------------------------------------------------------------------
#
def get_signal_by_name(signame):
    """
    Translate a signal name into the respective signal number.  If `signame` is
    not found to be a valid signal name, this method will raise a `KeyError`
    exception.  Lookup is case insensitive.
    """

    table = {'abrt'    : signal.SIGABRT,
             'alrm'    : signal.SIGALRM,
             'bus'     : signal.SIGBUS,
             'chld'    : signal.SIGCHLD,
           # 'cld'     : signal.SIGCLD,
             'cont'    : signal.SIGCONT,
             'fpe'     : signal.SIGFPE,
             'hup'     : signal.SIGHUP,
             'ill'     : signal.SIGILL,
             'int'     : signal.SIGINT,
             'io'      : signal.SIGIO,
             'iot'     : signal.SIGIOT,
             'kill'    : signal.SIGKILL,
             'pipe'    : signal.SIGPIPE,
           # 'poll'    : signal.SIGPOLL,
             'prof'    : signal.SIGPROF,
           # 'pwr'     : signal.SIGPWR,
             'quit'    : signal.SIGQUIT,
           # 'rtmax'   : signal.SIGRTMAX,
           # 'rtmin'   : signal.SIGRTMIN,
             'segv'    : signal.SIGSEGV,
             'stop'    : signal.SIGSTOP,
             'sys'     : signal.SIGSYS,
             'term'    : signal.SIGTERM,
             'trap'    : signal.SIGTRAP,
             'tstp'    : signal.SIGTSTP,
             'ttin'    : signal.SIGTTIN,
             'ttou'    : signal.SIGTTOU,
             'urg'     : signal.SIGURG,
             'usr1'    : signal.SIGUSR1,
             'usr2'    : signal.SIGUSR2,
             'vtalrm'  : signal.SIGVTALRM,
             'winch'   : signal.SIGWINCH,
             'xcpu'    : signal.SIGXCPU,
             'xfsz'    : signal.SIGXFSZ,
             }

    return table[signame.lower()]


# ------------------------------------------------------------------------------
#
class ThreadExit(SystemExit):
    pass


class SignalRaised(SystemExit):

    def __init__(self, msg, signum=None):
        if signum:
            msg = '%s [signal: %s]' % (msg, signum)
        SystemExit.__init__(self, msg)


# ------------------------------------------------------------------------------
#
def raise_in_thread(e=None, tname=None, tident=None):
    """
    This method uses an internal Python function to inject an exception 'e'
    into any given thread.  That thread can be specified by its name ('tname')
    or thread id ('tid').  If not specified, the exception is sent to the
    MainThread.

    The target thread will receive the exception with some delay.  More
    specifically, it needs to call up to 100 op codes before the exception
    is evaluated and raised.  The thread interruption can thus be delayed
    significantly, like when the thread sleeps.

    The default exception raised is 'radical.utils.ThreadExit' which inherits
    from 'SystemExit'.

    NOTE: this is not reliable: the exception is not raised immediately, but is
          *scheduled* for raising at some point, ie. in general after about 100
          opcodes (`sys.getcheckinterval()`).  Depending on when exactly the
          exception is finally raised, the interpreter might silently swallow
          it, if that happens in a generic try/except clause.  Those exist in
          the Python core, even if discouraged by some PEP or the other.

          See https://bugs.python.org/issue1779233

    NOTE: we can only raise exception *types*, not exception *instances*
          See https://bugs.python.org/issue1538556

    Example::

        def sub():
            time.sleep(0.1)
            ru.raise_in_thread()

        try:
            t = mt.Thread(target=sub)
            t.start()

            while True:
                time.sleep(0.01)

        except ru.ThreadExit:  print 'thread exit'
        except Exception as e: print 'except: %s' % e
        except SystemExit:     print 'exit'
        else:                  print 'unexcepted'
        finally:               t.join()
    """

    if not tident:
        if not tname:
            tname = 'MainThread'

        for th in mt.enumerate():
            if tname  == th.name:
                tident = th.ident
                break

    if not tident:
        raise ValueError('no target thread given/found')

    if not e:
        e = ThreadExit

    self_thread = mt.current_thread()
    if self_thread.ident == tident:
        # if we are in the target thread, we simply raise the exception.
        # This specifically also applies to the main thread.
        # Alas, we don't have a decent message to use...
        raise e('raise_in_thread')

    else:
        # otherwise we inject the exception into the main thread's async
        # exception scheduler
        ctypes.pythonapi.PyThreadState_SetAsyncExc(ctypes.c_long(tident),
                                                   ctypes.py_object(e))


# ------------------------------------------------------------------------------

